#include <assert.h>
#include <math.h>
#include "Bitcount.h"
#include "Bitboard.h"
#include "Attack.h"
using namespace std;


// TODO: must be used to speed up move generation
Bitboard get_pinned(Board& board, Color side)
{
  // TODO: improve this method using BetweenBB + popcount ()

  int s = (side == WHITE) ? 0 : 8;
  int o = (side == WHITE) ? 8 : 0;
  // optimisation clues
  // -not pinned chessmen on the edges of the board
  // -start from king sq, acting like attackers
  // -no pinned piece when doubled in a row before attackers
  // -no need to compute pinned chessmen even if king is in check (to cover double checks)
  
 
  Bitboard bbk = board.piecesBB[s+KING];
  int sq=pop_1st_bit(&bbk);
  
  // potential pinned chessmen
  Bitboard potPinnedByRQ = (board.colorBB[side] & RMasks[sq]);
  Bitboard potPinnedByBQ = (board.colorBB[side] & BMasks[sq]);
  
  Bitboard pinnedByRQ = 0L;
  Bitboard pinnedByBQ = 0L;
  
  // potential attackers
  Bitboard potentialRQ = (board.piecesBB[o+ROOK]   | board.piecesBB[o+QUEEN]) & RMasks[sq];
  Bitboard potentialBQ = (board.piecesBB[o+BISHOP] | board.piecesBB[o+QUEEN]) & BMasks[sq];
  
  // TODO: improve this part using BetweenBB[64][64] and popcount == 1 (to detect pinned pieces)
  File col = file_of((Square)sq);
  Rank row = rank_of((Square)sq);
  
  // Rook & Queen
  
  Square sqStart = (Square)sq + DELTA_E;
  Square sqEnd = (Square)(8 * (row+1));
  pinScanUp(sqStart, sqEnd, DELTA_E, potPinnedByRQ, potentialRQ, pinnedByRQ);

  sqStart = (Square)sq + DELTA_W;
  sqEnd = (Square)(8 * row - 1);
  pinScanDown(sqStart, sqEnd, DELTA_W, potPinnedByRQ, potentialRQ, pinnedByRQ);

  sqStart = (Square)sq + DELTA_N;
  sqEnd = (Square)64;
  pinScanUp(sqStart, sqEnd, DELTA_N, potPinnedByRQ, potentialRQ, pinnedByRQ);

  sqStart = (Square)sq + DELTA_S;
  sqEnd = (Square)-1;
  pinScanDown(sqStart, sqEnd, DELTA_S, potPinnedByRQ, potentialRQ, pinnedByRQ);
  

  // Bishop & Queen
   
  sqStart = (Square)sq + DELTA_NE;
  sqEnd = (Square)min(64, sq + 9 * (8 - col));
  pinScanUp(sqStart, sqEnd, DELTA_NE, potPinnedByBQ, potentialBQ, pinnedByBQ);

  sqStart = (Square)sq + DELTA_SE;
  sqEnd = (Square)max(-1, sq - 7 * (8 - col));
  pinScanDown(sqStart, sqEnd, DELTA_SE, potPinnedByBQ, potentialBQ, pinnedByBQ);

  sqStart = (Square)sq + DELTA_SW;
  sqEnd = (Square)max(-1, sq - 9 * (col + 1));
  pinScanDown(sqStart, sqEnd, DELTA_SW, potPinnedByBQ, potentialBQ, pinnedByBQ);

  sqStart = (Square)sq + DELTA_NW;
  sqEnd = (Square)min(64, sq + 7 * (col + 1));
  pinScanUp(sqStart, sqEnd, DELTA_NW, potPinnedByBQ, potentialBQ, pinnedByBQ); 

  return (pinnedByRQ | pinnedByBQ) ;
}

inline void pinScanUp(Square sqStart, Square sqEnd, Square sqStep, Bitboard& potPinned, Bitboard& potAttackers, Bitboard& pinned)
{
  Bitboard pin = 0L;
  for (Square sq = sqStart; sq < sqEnd; sq += sqStep)
  {
    if(SquareBB[sq] & potPinned)
    {
      if(pin) return; // two aligned pieces = no pin
      pin = SquareBB[sq];
    }
    else if(SquareBB[sq] & potAttackers)
    {
      pinned |= pin; // real pin
      return;
    }
  }
}
inline void pinScanDown(Square sqStart, Square sqEnd, Square sqStep, Bitboard& potPinned, Bitboard& potAttackers, Bitboard& pinned)
{
  Bitboard pin = 0L;
  for (Square sq = sqStart; sq > sqEnd; sq += sqStep)
  {
    if(SquareBB[sq] & potPinned)
    {
      if(pin) return; // two aligned pieces = no pin
      pin = SquareBB[sq];
    }
    else if(SquareBB[sq] & potAttackers)
    {
      pinned |= pin; // real pin
      return;
    }
  }
}

/*
inline void scanUp(Square sqStart, Square sqEnd, Square sqStep, Bitboard& potPinned, Bitboard& potAttackers, Bitboard& pinned)
{
  Bitboard pin = 0L;
  for (Square sq = sqStart; sq < sqEnd; sq += sqStep)
  {
    if(SquareBB[sq] & potPinned)
    {
      if(pin) return; // two aligned pieces = no pin
      pin = SquareBB[sq];
    }
    else if(SquareBB[sq] & potAttackers)
    {
      pinned |= pin;
      return;
    }
  }
}

inline void scanDown(Square sqStart, Square sqEnd, Square sqStep, Bitboard& potPinned, Bitboard& potAttackers, Bitboard& pinned)
{
  Bitboard pin = 0L;
  for (Square sq = sqStart; sq > sqEnd; sq += sqStep)
  {
    if(SquareBB[sq] & potPinned)
    {
      if(pin) return; // two aligned pieces = no pin
      pin = SquareBB[sq];
    }
    else if(SquareBB[sq] & potAttackers)
    {
      pinned |= pin;
      return;
    }
  }
}
*/

// Get attacks from the given side to the opposite side
void get_attacks(Board& board, Color side, attack_t& attack)
{
  int s = (side == WHITE) ? 0 : 8;
  
  // new implementation include opponent pieces 
  // a mask must be used later in the code
  //Bitboard notSide= ~board.colorBB[side] ; // empty squares or occupied by current side 

  Bitboard pMob = 0L;
  Bitboard pAtt = 0L;
  Bitboard nAtt = 0L;
  Bitboard kAtt = 0L;
  Bitboard bAtt = 0L;
  Bitboard rAtt = 0L;
  Bitboard qAtt = 0L;
  
  Square sq;
  Bitboard bb;

  // opponent's pawns
  for( bb = board.piecesBB[s+PAWN]; bb ; )
  {
    sq=pop_1st_bit(&bb);
    pAtt |= (StepAttacksBB[s+PAWN][sq]);
  }
  
  // opponent's knights
  for(bb = board.piecesBB[s+KNIGHT]; bb ; )
  {
    sq=pop_1st_bit(&bb);
    nAtt |= (StepAttacksBB[s+KNIGHT][sq]);
  }
  
  // opponent's king
  bb = board.piecesBB[s+KING];
  sq=pop_1st_bit(&bb);
  kAtt = (StepAttacksBB[s+KING][sq]);
  //print(att);
  
  
  // MORE COMPLICATED FOR SLIDERS
  // opponent's bishops
  for(bb = board.piecesBB[s+BISHOP]; bb ; )
  {
    sq=pop_1st_bit(&bb);
    bAtt |= get_bishop_attacks(board,sq);
  }
 
  // opponent's rooks
  for(bb = board.piecesBB[s+ROOK]; bb ; )
  {
    sq=pop_1st_bit(&bb);
    rAtt |= get_rook_attacks(board,sq);
  }
  
  // opponent's queens (=R+B)
  for(bb = board.piecesBB[s+QUEEN]; bb ; )
  {
    sq=pop_1st_bit(&bb);
    qAtt |= get_bishop_attacks(board,sq);
    qAtt |= get_rook_attacks(board,sq);    
  }  
 
  //board.DebugDisplay();
  //print(att);
  //assert(att);
  
  attack.allAttacks = ( pAtt | nAtt | kAtt | bAtt | rAtt | qAtt);
  
  attack.pawnMobility = pMob;
  attack.pawnAttacks = pAtt;
  attack.knightAttacks = nAtt;
  attack.kingAttacks = kAtt;
  attack.bishopAttacks = bAtt;
  attack.rookAttacks = rAtt;
  attack.queenAttacks = qAtt;
}

bool IsInCheck(Board& board, Color side, attack_t& attack)
{ 
  Piece king = (side == WHITE) ? W_KING : B_KING; 
  return (board.piecesBB[king] & attack.allAttacks);
}

Bitboard get_bishop_attacks(Board& board, Square& square)
{
  // draft attack bb before scan
  // TODO: use bitboards tricks for better perfs
  Bitboard att = 0L;
  
  File col = file_of(square);
  //Rank row = rank_of(square); // NOT USED
  
  // Up-right moves
  Square sqStart = square + DELTA_NE;
  Square sqEnd = (Square)min(64, square + 9 * (8 - col));
  for (Square sq = sqStart; sq < sqEnd; sq += DELTA_NE)
  {
    att |= SquareBB[sq];
    if(SquareBB[sq] & ~board.emptyBB)
      break;
  }
  
  // Down-right moves
  sqStart = square + DELTA_SE;
  sqEnd = (Square)max(-1, square - 7 * (8 - col));
  for (Square sq = sqStart; sq > sqEnd ; sq += DELTA_SE)
  {
    att |= SquareBB[sq];
    if(SquareBB[sq] & ~board.emptyBB)
      break;
  }
  
  // Down-left moves
  sqStart = square + DELTA_SW;
  sqEnd = (Square)max(-1, square - 9 * (col + 1));
  for (Square sq = sqStart ; sq > sqEnd; sq += DELTA_SW)
  {
    att |= SquareBB[sq];
    if(SquareBB[sq] & ~board.emptyBB) 
      break;
  }
  
  // Up-left moves
  sqStart = square + DELTA_NW;
  sqEnd = (Square)min(64, square + 7 * (col + 1));
  for (Square sq = sqStart; sq < sqEnd; sq += DELTA_NW)
  {
    att |= SquareBB[sq];
    if(SquareBB[sq] & ~board.emptyBB) 
      break;
  }
  
  return att;
}



Bitboard get_rook_attacks(Board& board, Square& square)
{
  Bitboard att = 0L;
  
  Rank row = rank_of(square);
  
  // Horizontal moves right
  Square sqStart = square + DELTA_E;
  Square sqEnd = (Square)(8 * (row+1));
  for (Square sq = sqStart; sq < sqEnd; sq++)
  {
    att |= SquareBB[sq];
    if(SquareBB[sq] & ~board.emptyBB) 
      break;
  }
  
  // Horizontal moves left
  sqStart = square + DELTA_W;
  sqEnd = (Square)(8 * row - 1);
  for (Square sq = sqStart; sq > sqEnd; sq--)
  {
    att |= SquareBB[sq];
    if(SquareBB[sq] & ~board.emptyBB) 
      break;
  }
  
  // Vertical moves up
  sqStart = (Square)(square + 8);
  sqEnd = (Square)64;
  for (Square sq = sqStart; sq < sqEnd; sq+=DELTA_N)
  {
    att |= SquareBB[sq];
    if(SquareBB[sq] & ~board.emptyBB) 
      break;
  }
  
  // Vertical moves down
  sqStart = square + DELTA_S;
  sqEnd = (Square)-1;
  for (Square sq = sqStart; sq > sqEnd; sq+=DELTA_S)
  {
    att |= SquareBB[sq];
    if(SquareBB[sq] & ~board.emptyBB) 
      break;
  }

  return att;
}
