#include <stdio.h>
#include <string.h>
#include <iostream>
#include <assert.h>
#include "Move.h"
#include "Board.h"
#include "Types.h"
#include "Bitboard.h"
#include "Bitcount.h"



bool is_correct_domove(Board* b , Move m)
{
  bool result = true;
  
  // after do_move, from_sq is empty
  if(b->occupiedBB & SquareBB[from_sq(m)])
  {
    std::cout <<  "from_sq=" << square_to_string(from_sq(m)) << " is not empty" << std::endl;
    result = false;
  }
  // after do_move, to_sq is no more empty
  if(b->emptyBB & SquareBB[to_sq(m)])
  {
    std::cout <<  "to_sq=" << square_to_string(to_sq(m)) << " is not occupied" << std::endl;
    result = false;
  }  
  
  if(result == false) 
  {
    std::cout <<  "problem on do move " << move_to_string(m) << std::endl;   
    b->DebugDisplay();
  }
  
  return result;
}



bool is_correct_undomove(Board* b , Move m)
{
  bool result = true;
  // after undo_move, from_sq is no more empty
  if(b->emptyBB & SquareBB[from_sq(m)])
  {
    std::cout <<  "from_sq=" << square_to_string(from_sq(m)) << " is not occupied" << std::endl;
    result = false;
  }
  // after undo_move, to_sq can be empty or opponent
  if(SquareBB[to_sq(m)] & b->colorBB[color_of(b->pieces[from_sq(m)])])
  {
    std::cout <<  "to_sq=" << square_to_string(to_sq(m)) << " is still occupied by player" << std::endl;
    result = false;
  } 
  
  if(result == false) 
  {
    std::cout <<  "problem on undo move " << move_to_string(m) << std::endl;   
    b->DebugDisplay();
  }
  
  return result;
}


bool board_is_correct(Board* b, std::string message)
{
  bool result = true;
  // check emptyBB/colorBB
  if(b->emptyBB & b->colorBB[WHITE])
  {
    std::cout <<  "conflict between emptyBB & colorBB[WHITE]" << std::endl;
    result = false;
  }
  if(b->emptyBB & b->colorBB[BLACK])
  {
    std::cout <<  "conflict between emptyBB & colorBB[BLACK]" << std::endl;
    result = false; 
  }
  if(b->colorBB[WHITE] & b->colorBB[BLACK])
  {
    std::cout <<  "conflict between colorBB[WHITE] & colorBB[BLACK]" << std::endl;
    result = false; 
  }
  
  // TODO: check each bitboard 
  
  // check white-white partition
  for(int i=W_PAWN;i<W_KING;i++)
    for(int j=(i+1);j<=W_KING;j++)
    {
      if(b->piecesBB[i] & b->piecesBB[j])
      {
        std::cout <<  "conflict between " << "-PNBRQK??pnbrqk -"[i] << " & " << "-PNBRQK??pnbrqk -"[j] << std::endl;
        result = false;
      }
    }
  // check black-black partition
  for(int i=B_PAWN;i<B_KING;i++)
    for(int j=(i+1);j<=B_KING;j++)
    {
      if(b->piecesBB[i] & b->piecesBB[j])
      {
        std::cout <<  "conflict between " << "-PNBRQK??pnbrqk -"[i] << " & " << "-PNBRQK??pnbrqk -"[j] << std::endl;
        result = false;
      }
    }
  // check white-black partition
  for(int i=W_PAWN;i<=W_KING;i++)
    for(int j=B_PAWN;j<=B_KING;j++)
    {
      if(b->piecesBB[i] & b->piecesBB[j])
      {
        std::cout <<  "conflict between " << "-PNBRQK??pnbrqk -"[i] << " & " << "-PNBRQK??pnbrqk -"[j] << std::endl;
        result = false;
      }
    }
  
  if(result == false) 
  {
    std::cout <<  message << std::endl;  
    b->DebugDisplay();
  }
  
  return result;
}


int count_pieces(Board* b)
{
  Bitboard notEmpty = ~b->emptyBB;
  return popcount<Full>(notEmpty);
}

Board::Board()
{
}

Board::~Board()
{
}

void Board::Init()
{
	this->Clear();
}

/*
 * Create a new board by copy
 */
void Board::Copy(Board & b_src)
{
	memcpy(this->pieces,b_src.pieces,64);
	memcpy(this->piecesBB,b_src.piecesBB,PIECEBB_SIZE);
	this->emptyBB = b_src.emptyBB;
	this->occupiedBB = b_src.occupiedBB;
	this->colorBB[WHITE] = b_src.colorBB[WHITE];
	this->colorBB[BLACK] = b_src.colorBB[BLACK];
}


void Board::Clear()
{
	for (int i = 0; i < 64; i++)
	{
		this->pieces[i] = NO_PIECE;
	}
	this->emptyBB = 0xffffffffffffffffULL;
	this->occupiedBB = 0;
  
	for (int i = 0; i < PIECEBB_SIZE; i++)
		this->piecesBB[i] = 0;
  
	this->colorBB[WHITE] = this->colorBB[BLACK] = 0;
}


void Board::ClearSquare(int square)
{
  //std::cout <<  "clear square " << square << std::endl;
  
	this->pieces[square] = NO_PIECE;
	this->emptyBB |= SquareBB[square];
	this->occupiedBB &= ~(SquareBB[square]);
  
	this->colorBB[WHITE] &= ~(SquareBB[square]);
	this->colorBB[BLACK] &= ~(SquareBB[square]);
  for(int i=0 ; i<PIECEBB_SIZE ; i++) this->piecesBB[i] &= ~(SquareBB[square]);
}

void Board::SetPiece(Piece piece, Square square)
{
	assert((square >= 0) && (square <= 63));
	assert(piece != NO_PIECE);
  //std::cout <<  "set piece on square " << square << std::endl;
  
	this->pieces[square] = piece;
  for(int i=0;i<PIECEBB_SIZE;i++) this->piecesBB[i] &= ~(SquareBB[square]);
	this->piecesBB[(int)piece] |= SquareBB[square];
	this->occupiedBB |= SquareBB[square];
	this->emptyBB &= ~(SquareBB[square]);
	this->colorBB[color_of(piece)] |= SquareBB[square];
	
	assert(this->piecesBB[(int)piece]);
	assert(this->occupiedBB);
	assert(this->colorBB[color_of(piece)]);
}


Piece Board::GetPiece(Square square)
{
	assert((square >= 0) && (square <= 63)); 
	return this->pieces[square];
}

// TODO: simplify and factorize some parts of this method
// Should be inline or not ?
void Board::DoMove(Move move)
{
  int nbPiecesBeforeMove = 0;
	// TODO: uncomment this part
	//assert(piecesBB[(int) W_KING] != 0);
	//assert(piecesBB[(int) B_KING] != 0);
  //std::cout <<  "check before Board::DoMove " << move_to_string(move) << std::endl;
	//this->DebugDisplay();
	
	assert(from_sq(move) != to_sq(move));
  assert(board_is_correct(this,"check before Board::DoMove "));
  assert(nbPiecesBeforeMove = count_pieces(this));
  
  
//  if(move == moveToTrack) {
//   std::cout <<  "before move to track" << std::endl;
//   this->DebugDisplay();
//   }
  
  
	Piece piece = this->pieces[from_sq(move)];
	Piece cPiece = this->pieces[to_sq(move)];
  
	this->pieces[to_sq(move)] = this->pieces[from_sq(move)];
	this->pieces[from_sq(move)] =  NO_PIECE;
  
	Bitboard fromBB   = SquareBB[from_sq(move)];
	Bitboard toBB     = SquareBB[to_sq(move)];
	Bitboard fromToBB = fromBB ^ toBB; // |+
	this->piecesBB[(int)piece]      ^= fromToBB;   // update piece bitboard
	this->colorBB[color_of(piece)]  ^= fromToBB;   // update white or black color bitboard
	if (cPiece ==  NO_PIECE)
	{
		// quiet move
		this->occupiedBB  ^= fromToBB;   // update occupied ...
		this->emptyBB     ^= fromToBB;   // ... and empty bitboard
	}
	else
	{
		// capture
		this->piecesBB[(int)cPiece]     ^= toBB;   // reset the captured piece
		this->colorBB[color_of(cPiece)] ^= toBB;   // update color bitboard by captured piece
		this->occupiedBB                ^= fromBB; // update occupied, only from becomes empty
		this->emptyBB                   ^= fromBB; // update empty bitboard
	}
  
  if(is_enpassant(move))
  {
    // TODO: improve en-passant sq calculation
    if(color_of(piece)==WHITE)
    {
      Square eps = to_sq(move) + DELTA_S;
      Bitboard epBB = toBB >> 8 ;
      pieces[eps]        = NO_PIECE;
      occupiedBB        ^= epBB; // remove from occupied
      emptyBB           ^= epBB; // add to empty
      piecesBB[B_PAWN]  ^= epBB;
      colorBB[BLACK]    ^= epBB;
    }
    else
    {
      Square eps = to_sq(move) + DELTA_N;   
      Bitboard epBB = toBB << 8 ;
      pieces[eps]        = NO_PIECE;
      occupiedBB        ^= epBB; // remove from occupied
      emptyBB           ^= epBB; // add to empty
      piecesBB[W_PAWN]  ^= epBB;
      colorBB[WHITE]    ^= epBB;
    }
  }
  // manage promotion (generate wrong bitboard status)
  else if(is_promotion(move))
  {
    piecesBB[piece]     &= ~toBB;      // remove pawn
    Piece newPiece       = piece - W_PAWN + (Piece)promotion_type(move); // compute new piece
    piecesBB[newPiece]  |= toBB;    // add promotion piece
    pieces[to_sq(move)]  = newPiece;
  }
  // Castleling extra-move (Rook)
  else if(move == make_castle(SQ_E1,SQ_G1))
  {
    pieces[SQ_H1] = NO_PIECE;
    pieces[SQ_F1] = W_ROOK;
    Bitboard fromToRookBB = (SquareBB[SQ_H1] | SquareBB[SQ_F1]);
    piecesBB[W_ROOK]    ^= fromToRookBB; 
    emptyBB             ^= fromToRookBB;
    occupiedBB          ^= fromToRookBB;
    colorBB[WHITE]      ^= fromToRookBB;
  }
  else if(move == make_castle(SQ_E1,SQ_C1))
  {
    pieces[SQ_A1]       = NO_PIECE;
    pieces[SQ_D1]       = W_ROOK;
    Bitboard fromToRookBB = (SquareBB[SQ_A1] | SquareBB[SQ_D1]); 
    piecesBB[W_ROOK]    ^= fromToRookBB; 
    emptyBB             ^= fromToRookBB;
    occupiedBB          ^= fromToRookBB;
    colorBB[WHITE]      ^= fromToRookBB;
  }
  else if(move == make_castle(SQ_E8,SQ_G8))
  {
    pieces[SQ_H8] = NO_PIECE;
    pieces[SQ_F8] = B_ROOK;
    Bitboard fromToRookBB = (SquareBB[SQ_H8] | SquareBB[SQ_F8]);
    piecesBB[B_ROOK]    ^= fromToRookBB;
    emptyBB             ^= fromToRookBB;
    occupiedBB          ^= fromToRookBB;
    colorBB[BLACK]      ^= fromToRookBB;
  }
  else if(move == make_castle(SQ_E8,SQ_C8))
  {
    pieces[SQ_A8] = NO_PIECE;
    pieces[SQ_D8] = B_ROOK;
    Bitboard fromToRookBB = (SquareBB[SQ_A8] | SquareBB[SQ_D8]); 
    piecesBB[B_ROOK]    ^= fromToRookBB;
    emptyBB             ^= fromToRookBB;
    occupiedBB          ^= fromToRookBB;
    colorBB[BLACK]      ^= fromToRookBB;
  }
  
  /*
   if(is_promotion(move)) {
   std::cout <<  "DO after promotion" << std::endl;
   this->DebugDisplay();
   }
   */
  //std::cout <<  "check after Board::DoMove " << move_to_string(move) << std::endl;
  //this->DebugDisplay();
  
  // TODO: regroup special move modifiers from Posiction.cpp to Board.cpp
  assert(board_is_correct(this,"check after Board::DoMove "+move_to_string(move)));
  assert(count_pieces(this) <= nbPiecesBeforeMove);
  assert(is_correct_domove(this,move));
}

// Should be inline or not ?
void Board::UndoMove(Move move, Piece piece, Piece cPiece)
{
  int nbPiecesBeforeUnMove = 0;
	//// TODO: uncomment this part
	//assert(piecesBB[(int) W_KING] != 0);
	//assert(piecesBB[(int) B_KING] != 0);
	assert(from_sq(move) != to_sq(move));
  assert(board_is_correct(this, "check before Board::UndoMove"));
  assert(nbPiecesBeforeUnMove = count_pieces(this));
  
  /* 
   if(is_promotion(move)) {
   std::cout <<  "UNDO after promotion" << std::endl;
   this->DebugDisplay();
   }
   */
	//Board::DebugDisplay(b);
  
	//Piece piece = pieces[to_sq(move)];
  
	this->pieces[from_sq(move)] = piece;
	this->pieces[to_sq(move)]   = cPiece;
  
	Bitboard fromBB   = SquareBB[from_sq(move)];
	Bitboard toBB     = SquareBB[to_sq(move)];
	Bitboard fromToBB = fromBB ^ toBB; // |+
  
  
  // manage promotion (generate wrong bitboard status)
  if(is_promotion(move))
  {
    piecesBB[piece]    |= toBB;      // add temp pawn at dest square
    Piece newPiece      = piece - W_PAWN + (Piece)promotion_type(move); // compute new piece
    piecesBB[newPiece] &= ~toBB;    // remove promotion piece
  }
  
	this->piecesBB[(int)piece]      ^= fromToBB;   // update piece bitboard
	this->colorBB[color_of(piece)]  ^= fromToBB;   // update white or black color bitboard
	if (cPiece ==  NO_PIECE)
	{
		// quiet move
		this->occupiedBB  ^= fromToBB;   // update occupied ...
		this->emptyBB     ^= fromToBB;   // ... and empty bitboard
  }
	else
	{
		// capture
		this->piecesBB[(int)cPiece]     ^= toBB;    // reset the captured piece
		this->colorBB[color_of(cPiece)] ^= toBB;    // update white or black color bitboard
		this->occupiedBB                ^= fromBB;  // update occupied, only from becomes empty
		this->emptyBB                   ^= fromBB;  // update empty bitboard
	}
  if(is_enpassant(move))
  {
    // TODO: improve en-passant sq calculation
    if(color_of(piece)==WHITE)
    {
      Square eps = to_sq(move) + DELTA_S;
      Bitboard epBB = toBB >> 8 ;
      pieces[eps] = B_PAWN;
      occupiedBB        ^= epBB;     
      emptyBB           ^= epBB;      
      piecesBB[B_PAWN]  ^= epBB;
      colorBB[BLACK]    ^= epBB;
    }
    else
    {
      Square eps = to_sq(move) + DELTA_N;   
      Bitboard epBB = toBB << 8 ;
      pieces[eps] = W_PAWN;
      occupiedBB        ^= epBB;     
      emptyBB           ^= epBB;      
      piecesBB[W_PAWN]  ^= epBB;
      colorBB[WHITE]    ^= epBB;
    }
  }
  // Castleling extra-move (Rook)
  else if(move == make_castle(SQ_E1,SQ_G1))
  {
    pieces[SQ_H1] = W_ROOK;
    pieces[SQ_F1] = NO_PIECE;
    Bitboard fromToRookBB = (SquareBB[SQ_H1] | SquareBB[SQ_F1]);
    piecesBB[W_ROOK]      ^= fromToRookBB; 
    emptyBB               ^= fromToRookBB;
    occupiedBB            ^= fromToRookBB;
    colorBB[WHITE]        ^= fromToRookBB;
  }
  else if(move == make_castle(SQ_E1,SQ_C1))
  {
    pieces[SQ_A1] = W_ROOK;
    pieces[SQ_D1] = NO_PIECE;
    Bitboard fromToRookBB = (SquareBB[SQ_A1] | SquareBB[SQ_D1]); 
    piecesBB[W_ROOK]      ^= fromToRookBB; 
    emptyBB               ^= fromToRookBB;
    occupiedBB            ^= fromToRookBB;
    colorBB[WHITE]        ^= fromToRookBB;
  }
  else if(move == make_castle(SQ_E8,SQ_G8))
  {
    pieces[SQ_H8] = B_ROOK;
    pieces[SQ_F8] = NO_PIECE;
    Bitboard fromToRookBB = (SquareBB[SQ_H8] | SquareBB[SQ_F8]);
    piecesBB[B_ROOK]      ^= fromToRookBB;
    emptyBB               ^= fromToRookBB;
    occupiedBB            ^= fromToRookBB;
    colorBB[BLACK]        ^= fromToRookBB;
  }
  else if(move == make_castle(SQ_E8,SQ_C8))
  {
    pieces[SQ_A8] = B_ROOK;
    pieces[SQ_D8] = NO_PIECE;
    Bitboard fromToRookBB = (SquareBB[SQ_A8] | SquareBB[SQ_D8]); 
    piecesBB[B_ROOK]      ^= fromToRookBB;
    emptyBB               ^= fromToRookBB;
    occupiedBB            ^= fromToRookBB;
    colorBB[BLACK]        ^= fromToRookBB;
  }    
  
  
  /*
   if(is_promotion(move)) {
   std::cout <<  "UNDO before promotion" << std::endl;
   this->DebugDisplay();
   }
   */
  assert(board_is_correct(this,"check after Board::UndoMove"));
  assert(nbPiecesBeforeUnMove = count_pieces(this));
  assert(is_correct_undomove(this,move));
}


bool Board::IsSquareEmpty(Square square)
{
	//return (this->pieces[square] == NO_PIECE);
  return (this->emptyBB & SquareBB[square]);
}

bool Board::IsEmptyOrColor(Square square, Color color)
{
	return SquareBB[square] & (emptyBB | colorBB[color]);
}

// FOR TESTING ONLY
bool Board::IsEqual(Board & b)
{
	for (int i = 0; i < PIECEBB_SIZE; i++) 
		if (this->piecesBB[i] != b.piecesBB[i]) 
			return false;
  
  // board should give the same result as for BB (double check)
	for (int i = 0; i < 64; i++) 
		if (this->pieces[i] != b.pieces[i]) 
			return false;
  
  for (int i = WHITE; i <= BLACK; i++) 
		if (this->colorBB[i] != b.colorBB[i]) 
			return false;
  
  if(this->emptyBB != b.emptyBB)
    return false;
  if(this->occupiedBB != b.occupiedBB)
    return false;
  
  return true;
}


void Board::DebugDisplay()
{
	// Board display
	// Useful to detect an incorrect position
	std::cout <<  "BOARD" << std::endl;
	for (int i = 7; i >= 0; i--)
	{
		std::cout <<  " +---+---+---+---+---+---+---+---+" << std::endl;
		for (int j = 0; j < 8; j++)
			std::cout <<  " | " << PieceToChar(this->pieces[i * 8 + j]);
		std::cout <<  " |" << std::endl;
	}
	std::cout <<  " +---+---+---+---+---+---+---+---+" << std::endl;
  
  /*
	// bitboards debug display 
	// Useful to detect an incorrect bitboard usage
	std::cout <<  "BITBOARDS" << std::endl;
	for(int i=0;i<16;i++) {
		// displays only non-empty
		if(piecesBB[i]) {
			std::cout <<  "BITBOARD [" << "-PNBRQK??pnbrqk -"[i] << "]" <<std::endl;
			print(piecesBB[i]);
		}
	}
	std::cout <<  "BITBOARDS occupiedBB" << std::endl;
	print(occupiedBB);
	std::cout <<  "BITBOARDS emptyBB" << std::endl;
	print(emptyBB);
	std::cout <<  "BITBOARDS colorBB[WHITE]" << std::endl;
	print(colorBB[WHITE]);
	std::cout <<  "BITBOARDS colorBB[BLACK]" << std::endl;
	print(colorBB[BLACK]);
  */
}


//NO_PIECE = 16, // color_of(NO_PIECE) == NO_COLOR
//W_PAWN = 1, W_KNIGHT =  2, W_BISHOP =  3, W_ROOK =  4, W_QUEEN =  5, W_KING =  6,
//B_PAWN = 9, B_KNIGHT = 10, B_BISHOP = 11, B_ROOK = 12, B_QUEEN = 13, B_KING = 14
inline char Board::PieceToChar(Piece p)
{
	return " PNBRQK  pnbrqk  "[(int)p];
}


bool Board::IsCheck(Color side)
{
	assert(this->piecesBB[(int) W_KING] != 0);
	assert(this->piecesBB[(int) B_KING] != 0);
	Square square = first_1(this->piecesBB[make_piece(side,KING)]);
	return IsCheckOnSquare(square, (Color)(1-side));
}


// Current version uses bitboards for step attacks & indep. ray scans for slinding pieces
bool Board::IsCheckOnSquare(Square square, Color bySide)
{
	assert((square >= 0) && (square < 64));
  
  // BITBOARD PART (faster part first)
	Bitboard pawns = this->piecesBB[make_piece(bySide,PAWN)];
	if (StepAttacksBB[make_piece(BLACK-bySide, PAWN)][square] & pawns) 
		return true;
  
	Bitboard knights = this->piecesBB[make_piece(bySide,KNIGHT)];
	if (StepAttacksBB[make_piece(bySide, KNIGHT)][square] & knights) 
		return true;
  
	Bitboard king = this->piecesBB[make_piece(bySide,KING)];
	if (StepAttacksBB[make_piece(bySide, KING)][square] & king) 
		return true;
  
	// bishop&queen threat
	// NO BITBOARD PART
	Piece oQueen = (bySide == BLACK) ? B_QUEEN : W_QUEEN;
	Piece oRook = (bySide == BLACK) ? B_ROOK : W_ROOK;
	Piece oBish = (bySide == BLACK) ? B_BISHOP : W_BISHOP;
  
	File col = file_of(square);
  Rank row = rank_of(square);
  
	// Up-right moves
	Square sqStart = square + DELTA_NE;
	Square sqEnd = (Square)std::min(64, square + 9 * (8 - col));
	for (Square sq = sqStart; sq < sqEnd; sq += DELTA_NE)
	{
		Piece p = this->GetPiece(sq);
		if ((p == oBish) || (p == oQueen))
			return true;
		if (p != NO_PIECE)
			break;
	}
  
	// Down-right moves
	sqStart = square + DELTA_SE;
	sqEnd = (Square)std::max(-1, square - 7 * (8 - col));
	for (Square sq = sqStart; sq > sqEnd; sq += DELTA_SE)
	{
		Piece p = this->GetPiece(sq);
		if ((p == oBish) || (p == oQueen))
			return true;
		if (p != NO_PIECE)
			break;
	}
  
	// Down-left moves
	sqStart = square + DELTA_SW;
	sqEnd = (Square)std::max(-1, square - 9 * (col + 1));
	for (Square sq = sqStart; sq > sqEnd; sq += DELTA_SW)
	{
		Piece p = this->GetPiece(sq);
		if ((p == oBish) || (p == oQueen))
			return true;
		if (p != NO_PIECE)
			break;
	}
  
	// Up-left moves
	sqStart = square + DELTA_NW;
	sqEnd = (Square)std::min(64, square + 7 * (col + 1));
	for (Square sq = sqStart; sq < sqEnd; sq += DELTA_NW)
	{
		Piece p = this->GetPiece(sq);
		if ((p == oBish) || (p == oQueen))
			return true;
		if (p != NO_PIECE)
			break;
	}
  
  
	// rook&queen threat
	// Horizontal moves right
	sqStart = square + DELTA_E;
	sqEnd = (Square)(8 * (row + 1));
	for (Square sq = sqStart; sq < sqEnd; sq++)
	{
		Piece p = this->GetPiece(sq);
		if ((p == oRook) || (p == oQueen))
			return true;
		if (p != NO_PIECE)
			break;
	}
  
	// Horizontal moves left
	sqStart = square + DELTA_W;
	sqEnd = (Square)(8 * row - 1);
	for (Square sq = sqStart; sq > sqEnd; sq--)
	{
		Piece p = this->GetPiece(sq);
		if ((p == oRook) || (p == oQueen))
			return true;
		if (p != NO_PIECE)
			break;
	}
  
	// Vertical moves up
	sqStart = square + DELTA_N;
	sqEnd = (Square)64;
	for (Square sq = sqStart; sq < sqEnd; sq+=DELTA_N)
	{
		Piece p = this->GetPiece(sq);
		if ((p == oRook) || (p == oQueen))
			return true;
		if (p != NO_PIECE)
			break;
	}
	
	// Vertical moves down
	sqStart = square + DELTA_S;
	sqEnd = (Square)-1;
	for (Square sq = sqStart; sq > sqEnd; sq+=DELTA_S)
	{
		Piece p = this->GetPiece(sq);
		if ((p == oRook) || (p == oQueen))
			return true;
		if (p != NO_PIECE)
			break;
	}
  
	return false;
}

