#include <map>
#include <string>
#include <iostream>
#include "Move.h"
#include "DebugBench.h"
#include "Engine.h"
#include "Watch.h"
#include "MoveGen.h"
#include "Zobrist.h"

using std::cin;
using std::string;
using std::map;

#define STR_RESULT_OK "-- OK --";
#define STR_RESULT_KO "** KO **";

void DebugBench::LaunchBench()
{
	map<string,string> simplePositionsDB;
	map<string,string> mediumPositionsDB;
	map<string,string> complexPositionsDB;

	initSimplePositionsDB(simplePositionsDB);
	initMediumPositionsDB(mediumPositionsDB);
	initComplexPositionsDB(complexPositionsDB);

	std::cout <<  
	  "*****************************************\n"
		"** PERFORMANCE TEST SET                **\n"
		"*****************************************\n"
		" 1 - BenchMoveGeneration                 \n"
		" 2 - BenchTranspositionTable             \n"
		"*****************************************\n";

	std::cout <<  "Bench:";
	int choice;
	cin >> choice;

	switch (choice)
	{
	  case 1:
		  BenchMoveGeneration(simplePositionsDB);
		  break;
	  case 2:
		  BenchTranspositionTable(simplePositionsDB);
		  break;
	  default:
		  break;
	}
}


void DebugBench::BenchMoveGeneration(map<string,string> & positionsDB)
{

	const int MAXMGEN = 10000;
	Watch watch1,watch2,watch3;
	list_t legalMoves;
	Position pos;
	vector<double> pseudoLegMoveGet, legalMoveGet, legalMoveGen;

	for(map<string,string>::iterator item = positionsDB.begin(); item != positionsDB.end() ; item++)
	{
		string fen = (*item).first;
		pos.SetFromFEN(fen);

		// pseudo legal moves
		watch1.Start();
		for(int i=0; i<MAXMGEN;i++)
		{
			GetPossibleMoves(pos,legalMoves);
		}
		watch1.Stop();

		// legal moves
		watch2.Start();
		for(int i=0; i<MAXMGEN;i++)
		{
			GetLegalMoves(pos,legalMoves);
		}
		watch2.Stop();

		// legal moves
		watch3.Start();
		for(int i=0; i<MAXMGEN;i++)
		{
			//GenLegalMoves(pos,legalMoves);
		}
		watch3.Stop();

		pseudoLegMoveGet.push_back((double)watch1.GetElapsedTime()/MAXMGEN);
		legalMoveGet.push_back((double)watch2.GetElapsedTime()/MAXMGEN);
    legalMoveGen.push_back((double)watch3.GetElapsedTime()/MAXMGEN);
	}

	double meanPseudoLegMoveGet = Mean(pseudoLegMoveGet);
	double meanLegalMoveGet = Mean(legalMoveGet);
  double meanLegalMoveGen = Mean(legalMoveGen);
	
	std::cout <<  "********************************************"<< std::endl;
	std::cout <<  "* BenchPositionsDB                         *"<< std::endl;
	std::cout <<  "********************************************"<< std::endl;
	std::cout <<  "Nb of benched positions: " << positionsDB.size() << "x" << MAXMGEN  << "=" << positionsDB.size() * MAXMGEN << std::endl;
	std::cout <<  " ------------- Pseudo legal moves generation" << std::endl;
	std::cout <<  "time per position: " << (meanPseudoLegMoveGet*1000) << " microsec" << std::endl;
	std::cout <<  "Average speed:     " << (meanPseudoLegMoveGet?(int)(1/meanPseudoLegMoveGet):0) << " Knps" << std::endl;
	std::cout <<  " -------------------- Legal moves generation" << std::endl;
	std::cout <<  "time per position: " << (meanLegalMoveGet*1000) << " microsec" << std::endl;
	std::cout <<  "Average speed:     " << (meanLegalMoveGet?(int)(1/meanLegalMoveGet):0) << " Knps" << std::endl;
	std::cout <<  "********************************************"<< std::endl;
	std::cout <<  "time per position: " << (meanLegalMoveGen*1000) << " microsec" << std::endl;
	std::cout <<  "Average speed:     " << (meanLegalMoveGen?(int)(1/meanLegalMoveGen):0) << " Knps" << std::endl;
	std::cout <<  "********************************************"<< std::endl;

}

/*
void DebugBench::BenchPositionsDB(map<string,string> & positionsDB)
{
 
}
*/

void DebugBench::BenchTranspositionTable(map<string,string> & positionsDB)
{
  TranspositionTable TTable;
  const int MAXMGEN = 10000;
	Watch watch1,watch2,watch3;
	int posNb = (int)positionsDB.size();
	//Position positions[posNb];
	Position positions[256];
	
	// preparing data set
	int i = 0;
	for(map<string,string>::iterator item = positionsDB.begin(); item != positionsDB.end() ; item++)
	{
	  string fen = item->first;
	  positions[i].SetFromFEN(fen);
	  i++;
	}

  // Write bench
  watch2.Start();
  for(int ii=0; ii<MAXMGEN;ii++)
  for(int i=0;i<posNb;i++)
	{
    //Key key = book_key(positions[i]);
	  TTElem tte;
	  //tte.type = TT_EXACT;
    //tte.bestMove = MOVE_NULL;
    tte.depth = 9;
    //tte.score = 2000;
    tte.alpha = 4999; //all nodes
    tte.beta = 1000+i; //cut nodes
    if(tte.beta<=500000) tte.beta = 1;
    //tte.age = 1566648;
	  //TTable.Save(key,tte);
	}
	watch2.Stop();
	
	// Write bench
  watch1.Start();
  for(int i=0; i<MAXMGEN;i++)
	for(int i=0;i<posNb;i++)
	{
	  TTElem* tte;
    //Key key = book_key(positions[i]);
	  //tte = TTable.Find(key);
	  if(tte != NULL)
	  {
	    //std::cout <<  "not found" << std::endl;
	  }
	  else
	  {
	    //std::cout <<  "found" << std::endl;
	  }
	}
	watch1.Stop();

  // key calculation bench
  watch3.Start();
  for(int ii=0; ii<MAXMGEN;ii++)
  for(int i=0;i<posNb;i++)
	{
    //Key key = book_key(positions[i]);
	}
	watch3.Stop();

  size_t totalNbPos = MAXMGEN * positionsDB.size();
  
  uint64_t timeReadTT = watch1.GetElapsedTime();
  uint64_t timeWriteTT = watch2.GetElapsedTime();
  uint64_t timeKeyTT = watch3.GetElapsedTime();
  
  double meanReadTimePerPos = totalNbPos? (double)timeReadTT/totalNbPos : 0.0;
	double meanWriteTimePerPos = totalNbPos? (double)timeWriteTT/totalNbPos : 0.0;
  double meanKeyTimePerPos = totalNbPos? (double)timeKeyTT/totalNbPos : 0.0;
  
	double meanReadSpeed = timeReadTT? (double)totalNbPos/timeReadTT : 0.0;
	double meanWriteSpeed = timeWriteTT? (double)totalNbPos/timeWriteTT : 0.0;
  double meanKeySpeed = timeKeyTT? (double)totalNbPos/timeKeyTT : 0.0;

	std::cout <<  "********************************************"<< std::endl;
	std::cout <<  "* BenchTranspositionTable                  *"<< std::endl;
	std::cout <<  "********************************************"<< std::endl;
	std::cout <<  "Nb of benched positions: " << positionsDB.size() << "x" << MAXMGEN  << "=" << positionsDB.size() * MAXMGEN << std::endl;
	std::cout <<  " -------------------------------------- Read" << std::endl;
	std::cout <<  "time per position: " << (int)meanReadTimePerPos << " microsec" << std::endl;
	std::cout <<  "Average speed:     " << meanReadSpeed << " pos/ms" << std::endl;
	std::cout <<  " ------------------------------------- Write" << std::endl;
	std::cout <<  "time per position: " << (int)meanWriteTimePerPos << " microsec" << std::endl;
	std::cout <<  "Average speed:     " << meanWriteSpeed << " pos/ms" << std::endl;
	std::cout <<  " ------------------------------------- Key" << std::endl;
  std::cout <<  "time per position: " << (int)meanKeyTimePerPos << " microsec" << std::endl;
	std::cout <<  "Average speed:     " << meanKeySpeed << " pos/ms" << std::endl;
	std::cout <<  "********************************************"<< std::endl;
}


double DebugBench::Mean(vector<double> results)
{
	double sum = 0;
	for(vector<double>::iterator nb = results.begin(); nb != results.end(); nb++)
	{
	  //std::cout <<  "nb=" << (*nb) << std::endl;
		sum += (*nb);
	}
	double mean = (double)(sum / results.size());
	//std::cout <<  "*** mean=" << mean << std::endl;
	return mean;
}

void DebugBench::initComplexPositionsDB(map<string,string> & complexPositionsDB)
{
	// TODO: change this set for a more representative sample (ex: Nunn problems or middle game positions)
	complexPositionsDB["rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR w KQkq - 0 1"] = "start position (of course the most complex one)";
	//complexPositionsDB["k6K/8/8/8/8/8/PPPPPPPP/8 w - - 1 1"] = "pawn moves 1:a2a3 a2a4 b2b3 b2b4 c2c3 c2c4 d2d3 d2d4 e2e3 e2e4 f2f3 f2f4 g2g3 g2g4 h2h3 h2h4 h8g8 h8g7 h8h7"; //w pawn moves
	/*
	complexPositionsDB["k1K5/p7/8/8/8/8/8/8 b - - 1 1"] = "pawn + king:a7a6 a7a5"; //b basic pawn moves
	complexPositionsDB["rn2kbnr/2Nbpppp/8/2pP4/3P1B2/P7/P1P2PPP/R2QKB1R b KQkq - 1 1"] = "check - only one move 1:e8d8";
	complexPositionsDB["r3k2r/8/8/8/8/p6p/P6P/R3K3 w KQkq - 1 1"] = "casteling moves:a1b1 a1c1 a1d1 e1d1 e1d2 e1e2 e1f2 e1f1 e1c1";
	complexPositionsDB["rnbqkbnr/4pQ2/2pp3p/6B1/p1BPP3/2N2NP1/PPP2PP1/R3K2R b KQkq - 0 11"] = "check - only one move 2:e8d7";
	complexPositionsDB["K6k/8/8/8/8/8/8/8 b - - 0 1"] = "king moves:h8g8 h8g7 h8h7";
	complexPositionsDB["rB2kbbr/1p6/8/p6Q/2PPP3/8/PP3PPP/R4RK1 b - - 0 19"] = "blacks in check - 5 moves:h8h5 g8f7 e8e7 e8d7 e8d8";
	complexPositionsDB["4kb1r/4b3/8/7B/8/8/8/6K1 b k - 0 19"] = "Blacks in check - escape or take(3 moves):h8h5 e8d8 e8d7";
	complexPositionsDB["K1k5/8/8/8/3R4/8/8/8 w - - 0 1 "] = "tower moves 1:d4d1 d4d2 d4d3 d4d5 d4d6 d4d7 d4d8 d4a4 d4b4 d4c4 d4e4 d4f4 d4g4 d4h4 a8a7";
	complexPositionsDB["K1k5/8/3p4/8/1P1R2p1/3P4/8/8 w - - 0 1 "] = "tower moves 2:d4d5 d4d6 d4c4 d4e4 d4f4 d4g4 a8a7 b4b5";
	complexPositionsDB["K1k5/8/8/8/8/3B4/8/8 w - - 0 1 "] = "bishop moves 1:d3b1 d3c2 d3e4 d3f5 d3g6 d3h7 d3a6 d3b5 d3c4 d3e2 d3f1 a8a7";
	complexPositionsDB["K1k5/4B3/1B6/8/6B1/3B4/8/8 w - - 0 1 "] = "bishop moves 2:d3b1 d3c2 d3e4 d3f5 d3g6 d3h7 d3a6 d3b5 d3c4 d3e2 d3f1 g4h3 g4h5 g4d1 g4e2 g4f3 g4f5 g4e6 g4d7 g4c8 e7f8 e7d8 e7f6 e7g5 e7h4 e7d6 e7c5 e7b4 e7a3 b6a7 b6c7 b6d8 b6a5 b6c5 b6d4 b6e3 b6f2 b6g1 a8a7";
	complexPositionsDB["8/8/8/4N3/8/8/8/k6K w - - 10 10"] = "knight moves 1:e5f7 e5g6 e5g4 e5f3 e5d3 e5c4 e5c6 e5d7 h1h2 h1g2 h1g1"; // knight moves
	complexPositionsDB["k7/8/8/pP6/8/8/8/K7 w - a6 10 10"] = "en-passant + pawn push 1 (5 moves):a1a2 a1b2 a1b1 b5b6 b5a6"; // kings + en passant
	complexPositionsDB["k7/8/8/1Pp5/8/8/8/K7 w - c6 10 10"] = "en-passant + pawn push 2 (5 moves):a1a2 a1b2 a1b1 b5b6 b5c6"; // kings + en passant
*/
}

void DebugBench::initMediumPositionsDB(map<string,string> & mediumPositionsDB)
{
	// TODO: change this set for a more representative sample
	mediumPositionsDB["k6K/8/8/8/8/8/PPPPPPPP/8 w - - 1 1"] = "pawn moves 1:a2a3 a2a4 b2b3 b2b4 c2c3 c2c4 d2d3 d2d4 e2e3 e2e4 f2f3 f2f4 g2g3 g2g4 h2h3 h2h4 h8g8 h8g7 h8h7"; //w pawn moves
	/*
	mediumPositionsDB["k1K5/p7/8/8/8/8/8/8 b - - 1 1"] = "pawn + king:a7a6 a7a5"; //b basic pawn moves
	mediumPositionsDB["rn2kbnr/2Nbpppp/8/2pP4/3P1B2/P7/P1P2PPP/R2QKB1R b KQkq - 1 1"] = "check - only one move 1:e8d8";
	mediumPositionsDB["r3k2r/8/8/8/8/p6p/P6P/R3K3 w KQkq - 1 1"] = "casteling moves:a1b1 a1c1 a1d1 e1d1 e1d2 e1e2 e1f2 e1f1 e1c1";
	mediumPositionsDB["rnbqkbnr/4pQ2/2pp3p/6B1/p1BPP3/2N2NP1/PPP2PP1/R3K2R b KQkq - 0 11"] = "check - only one move 2:e8d7";
	mediumPositionsDB["K6k/8/8/8/8/8/8/8 b - - 0 1"] = "king moves:h8g8 h8g7 h8h7";
	mediumPositionsDB["rB2kbbr/1p6/8/p6Q/2PPP3/8/PP3PPP/R4RK1 b - - 0 19"] = "blacks in check - 5 moves:h8h5 g8f7 e8e7 e8d7 e8d8";
	mediumPositionsDB["4kb1r/4b3/8/7B/8/8/8/6K1 b k - 0 19"] = "Blacks in check - escape or take(3 moves):h8h5 e8d8 e8d7";
	mediumPositionsDB["K1k5/8/8/8/3R4/8/8/8 w - - 0 1 "] = "tower moves 1:d4d1 d4d2 d4d3 d4d5 d4d6 d4d7 d4d8 d4a4 d4b4 d4c4 d4e4 d4f4 d4g4 d4h4 a8a7";
	mediumPositionsDB["K1k5/8/3p4/8/1P1R2p1/3P4/8/8 w - - 0 1 "] = "tower moves 2:d4d5 d4d6 d4c4 d4e4 d4f4 d4g4 a8a7 b4b5";
	mediumPositionsDB["K1k5/8/8/8/8/3B4/8/8 w - - 0 1 "] = "bishop moves 1:d3b1 d3c2 d3e4 d3f5 d3g6 d3h7 d3a6 d3b5 d3c4 d3e2 d3f1 a8a7";
	mediumPositionsDB["K1k5/4B3/1B6/8/6B1/3B4/8/8 w - - 0 1 "] = "bishop moves 2:d3b1 d3c2 d3e4 d3f5 d3g6 d3h7 d3a6 d3b5 d3c4 d3e2 d3f1 g4h3 g4h5 g4d1 g4e2 g4f3 g4f5 g4e6 g4d7 g4c8 e7f8 e7d8 e7f6 e7g5 e7h4 e7d6 e7c5 e7b4 e7a3 b6a7 b6c7 b6d8 b6a5 b6c5 b6d4 b6e3 b6f2 b6g1 a8a7";
	mediumPositionsDB["8/8/8/4N3/8/8/8/k6K w - - 10 10"] = "knight moves 1:e5f7 e5g6 e5g4 e5f3 e5d3 e5c4 e5c6 e5d7 h1h2 h1g2 h1g1"; // knight moves
	//mediumPositionsDB["k7/8/8/pP6/8/8/8/K7 w - a6 10 10"] = "en-passant + pawn push 1 (5 moves):a1a2 a1b2 a1b1 b5b6 b5a6"; // kings + en passant
	//mediumPositionsDB["k7/8/8/1Pp5/8/8/8/K7 w - c6 10 10"] = "en-passant + pawn push 2 (5 moves):a1a2 a1b2 a1b1 b5b6 b5c6"; // kings + en passant
*/
}

void DebugBench::initSimplePositionsDB(map<string,string> & simplePositionsDB)
{
	// TODO: change this set for a more representative sample (ex: endgames with few pieces)
	simplePositionsDB["k6K/8/8/8/8/8/PPPPPPPP/8 w - - 1 1"] = "pawn moves 1:a2a3 a2a4 b2b3 b2b4 c2c3 c2c4 d2d3 d2d4 e2e3 e2e4 f2f3 f2f4 g2g3 g2g4 h2h3 h2h4 h8g8 h8g7 h8h7"; //w pawn moves
	simplePositionsDB["k1K5/p7/8/8/8/8/8/8 b - - 1 1"] = "pawn + king:a7a6 a7a5"; //b basic pawn moves
	simplePositionsDB["rn2kbnr/2Nbpppp/8/2pP4/3P1B2/P7/P1P2PPP/R2QKB1R b KQkq - 1 1"] = "check - only one move 1:e8d8";
	simplePositionsDB["r3k2r/8/8/8/8/p6p/P6P/R3K3 w KQkq - 1 1"] = "casteling moves:a1b1 a1c1 a1d1 e1d1 e1d2 e1e2 e1f2 e1f1 e1c1";
	simplePositionsDB["rnbqkbnr/4pQ2/2pp3p/6B1/p1BPP3/2N2NP1/PPP2PP1/R3K2R b KQkq - 0 11"] = "check - only one move 2:e8d7";
	simplePositionsDB["K6k/8/8/8/8/8/8/8 b - - 0 1"] = "king moves:h8g8 h8g7 h8h7";
	simplePositionsDB["rB2kbbr/1p6/8/p6Q/2PPP3/8/PP3PPP/R4RK1 b - - 0 19"] = "blacks in check - 5 moves:h8h5 g8f7 e8e7 e8d7 e8d8";
	simplePositionsDB["4kb1r/4b3/8/7B/8/8/8/6K1 b k - 0 19"] = "Blacks in check - escape or take(3 moves):h8h5 e8d8 e8d7";
	simplePositionsDB["K1k5/8/8/8/3R4/8/8/8 w - - 0 1 "] = "tower moves 1:d4d1 d4d2 d4d3 d4d5 d4d6 d4d7 d4d8 d4a4 d4b4 d4c4 d4e4 d4f4 d4g4 d4h4 a8a7";
	simplePositionsDB["K1k5/8/3p4/8/1P1R2p1/3P4/8/8 w - - 0 1 "] = "tower moves 2:d4d5 d4d6 d4c4 d4e4 d4f4 d4g4 a8a7 b4b5";
	simplePositionsDB["K1k5/8/8/8/8/3B4/8/8 w - - 0 1 "] = "bishop moves 1:d3b1 d3c2 d3e4 d3f5 d3g6 d3h7 d3a6 d3b5 d3c4 d3e2 d3f1 a8a7";
	simplePositionsDB["8/8/8/4N3/8/8/8/k6K w - - 10 10"] = "knight moves 1:e5f7 e5g6 e5g4 e5f3 e5d3 e5c4 e5c6 e5d7 h1h2 h1g2 h1g1"; // knight moves
	simplePositionsDB["k7/8/8/pP6/8/8/8/K7 w - a6 10 10"] = "en-passant + pawn push 1 (5 moves):a1a2 a1b2 a1b1 b5b6 b5a6"; // kings + en passant
	simplePositionsDB["k7/8/8/1Pp5/8/8/8/K7 w - c6 10 10"] = "en-passant + pawn push 2 (5 moves):a1a2 a1b2 a1b1 b5b6 b5c6"; // kings + en passant
	
	// KO for getLegalMoves
	//simplePositionsDB["K1k5/4B3/1B6/8/6B1/3B4/8/8 w - - 0 1 "] = "bishop moves 2:d3b1 d3c2 d3e4 d3f5 d3g6 d3h7 d3a6 d3b5 d3c4 d3e2 d3f1 g4h3 g4h5 g4d1 g4e2 g4f3 g4f5 g4e6 g4d7 g4c8 e7f8 e7d8 e7f6 e7g5 e7h4 e7d6 e7c5 e7b4 e7a3 b6a7 b6c7 b6d8 b6a5 b6c5 b6d4 b6e3 b6f2 b6g1 a8a7";
	
}


void DebugBench::BenchSetFenPerfs()
{
/*
	Watch wa,wb;
	int nbPieces = 0;

	for (int i = 0; i < 1000; i++)
	{
	string fen = getRandomFEN();
	ArrayBoard aPos = new ArrayBoard();
	BitBoard bPos = new BitBoard();

	wa.Start();
	aPos.SetFromFEN(fen);
	wa.Stop();
	wb.Start();
	bPos.SetFromFEN(fen);
	wb.Stop();
	nbPieces += GetNbPieces(aPos);
	}
	System.std::cout <<  "Perfs for std call of SetFenPerfs (32 pieces/board)");
	System.std::cout <<  "ArrayBoard : " + 32 * wa.ElapsedTicks / nbPieces + " ticks/pos");
	System.std::cout <<  "BitBoard   : " + 32 * wb.ElapsedTicks / nbPieces + " ticks/pos");
*/
}


void DebugBench::BenchGetLegalMovesPerfs()
{
/*
	Watch wa;// = new Watch();
	Watch wb;// = new Watch();
	int nbPieces = 0;

	for (int i = 0; i < 1000; i++)
	{
	string fen = getRandomFEN();
	Position p = new Position();
	p.SetFromFEN(fen);

	wa.Start();
	GetPossibleMoves(p);
	wa.Stop();
	wb.Start();
	GetLegalMoves(p);
	wb.Stop();
	nbPieces += GetNbPieces(p);
	}
	std::cout <<  "Perfs for std call of GetLegalMoves (32 pieces/board)");
	std::cout <<  "GetPossibleMoves : " + 32 * wa.ElapsedTicks / nbPieces + " ticks/pos  (total:" + wa.ElapsedTicks + "ticks)");
	std::cout <<  "GetLegalMoves   : " + 32 * wb.ElapsedTicks / nbPieces + " ticks/pos (total:" + wb.ElapsedTicks + "ticks)");
*/
}






