#include <stdio.h>
#include <assert.h>
#include <iostream>
#include <sstream>
#include <vector>
#include <list>
#include <math.h>
#include "Engine.h"
#include "Move.h"
#include "Types.h"
#include "Platform.h"
#include "Thread.h"
#include "Transposition.h"
#include "MoveGen.h"
#include "Eval.h"
#include "Zobrist.h"
#include "Search.h"

using namespace std;

// properties
// TODO: turn them into privates
bool debugMode;
bool useTrans;
Position currentPosition;
Watch watch;
unsigned long nodeMaxNumber;
unsigned long nbProcessedNodes;// total nb of processed node in a search
unsigned long nbCurrSearchTotalGenNodes;
unsigned long nbCurrSearchProcessedNodes;
int cutOffStatTable[20];
unsigned long sumNodeCutOffIndex; // sum of indexes where the cuttoff occurs in each node
int selectiveDepth;
BrainMode brainMode;
BrainStatus brainStatus;
int depthLevel;
unsigned long thinkDelay;
long movestogo;
//bool UCI_LimitStrength;
//int UCI_Elo;
bool UCI_AnalyseMode;
// MultiCut parameters
int M;// M is the number of moves to look at when checking for mc-prune.
int C;// C is the number of cutoffs to cause an mc-prune, C < M.
int R;// R is the search depth reduction for mc-prune searches.
//TODO: replace by time[Color][] and inc[Color][]
long wtime;
long btime;
long winc;
long binc;
bool isVerbose;
bool isSmartTiming;
bool isBlackbox; // no-info mode to play at 100% full strength
bool isSearchMove;
// TODO: revert use of vectors to old style structs
list_t moves;
Move bestMove;
int bestScore;
Move* bestPv;
//vector<MoveScorePv> scores; // TODO: merge with pvbuffer
Move pvbuffer[MAX_MOVES][64]; //pvbuffer[max move nb for depth 0][max depth];
map<Move,Move*> pvhash; 
TranspositionTable TTable;


//long history[2][64][64]; // history heuristic
long history[16][64]; // history heuristic

void clear_history()
{
  memset(history,  0, 16 * 64 * sizeof(Value));
}

void engine_init()
{
// MULTICUT PARAMETERS
  M = 6;// M is the number of moves to look at when checking for mc-prune.
  C = 3;// C is the number of cutoffs to cause an mc-prune, C < M.
  R = 2;// R is the search depth reduction for mc-prune searches.

  debugMode = false;
  useTrans = true; // no improvment noticed for now (use of TT must be inaccurate)

  isSearchMove = false;
  brainStatus = PAUSED;

  // TODO: move this at SetStartPosition()
  // init heuristic history
  clear_history();
  
  // TODO debug init = Blitz 2'+6"/ply - to be removed
  SetWtime(120000);
  SetWinc(6000);
  SetBtime(120000);
  SetBinc(6000);
  SetStartPosition();

  options_init();
}

void options_init()
{
  SetOption("Smart Timer","true");
  SetOption("Hash","16");
  SetOption("UCI_AnalyseMode","false");
  //SetOption("UCI_LimitStrength","false");
  //SetOption("UCI_Elo", "1200" );	
  
  SetOption("Queen weight" , "900");
  SetOption("Rook weight"  , "500");
  SetOption("Bishop weight", "325");
  SetOption("Knight weight", "325");
  SetOption("Pawn weight"  , "85");

  SetOption("King mobility"  , "100");
  SetOption("Queen mobility" , "100");
  SetOption("Rook mobility"  , "100");
  SetOption("Bishop mobility", "100");
  SetOption("Knight mobility", "100");
  SetOption("Pawn mobility"  , "100"); 

  SetOption("Pawn structure" , "5");
  
  SetOption("Blackbox", "false"); 
}

void Play(Move m){
  currentPosition.Play(m);
}

void Quit()
{
  brainStatus = QUITTING;
}


void DisplayScores(list_t& moveList, int depth)
{
  std::cout <<  "NEW MOVES EVALUATION -------------------- depth " << depth << std::endl;
  int mvCount = 0;
  for(int i=0; i<LIST_SIZE(moveList);i++)
  {
    Move move = LIST_MOVE(moveList,i);
    int score = LIST_VALUE(moveList,i);
    Move* pv = pvhash[move];
    std::cout <<  move_to_string(move) << "=";
    if( score < -VALUE_KNOWN_WIN || score > VALUE_KNOWN_WIN )
      std::cout <<  ScoreToString(score) << ' ';
    else
      std::cout <<  score << ' ';
	  
    std::cout <<  " pv: ";
    DisplayPv(pv, depth);
	  std::cout <<  std::endl;
    mvCount++;
  }
  std::cout <<  std::endl ;
  std::cout <<  "-------------------------------------------------" << std::endl;
}


// Sort moves depending on the final score
// Bubble sort - can be improved
// TODO: improve the sort method
void Sort(vector<MoveStack> & scores)
{
  int i;
  int j;
  MoveStack temp;
  int x = (int)scores.size();

  for (i = (x - 1); i >= 0; i--)
    {
      for (j = 1; j <= i; j++)
	    {
	      if (scores[j - 1].score < scores[j].score)
        {
          temp = scores[j - 1];
          scores[j - 1] = scores[j];
          scores[j] = temp;
        }
	    }
    }
}
/*
void Sort(vector<MoveScorePv> & scores)
{
  int i;
  int j;
  MoveScorePv temp;
  int x = (int)scores.size();

  for (i = (x - 1); i >= 0; i--)
    {
      for (j = 1; j <= i; j++)
	    {
	      if (scores[j - 1].score < scores[j].score)
        {
          temp = scores[j - 1];
          scores[j - 1] = scores[j];
          scores[j] = temp;
        }
	    }
    }
}
*/
void Sort(list_t& t)
{
  int i,j;
  Move tmove;
  int tvalue;  
  for (i = (t.size - 1); i >= 0; i--)
  {
    for (j = 1; j <= i; j++)
    {
      if (LIST_VALUE(t,j - 1) < LIST_VALUE(t,j))
      {
        tmove   = LIST_MOVE(t,j - 1);
        tvalue  = LIST_VALUE(t,j - 1);
        LIST_MOVE(t,j - 1)  = LIST_MOVE(t,j);
        LIST_VALUE(t,j - 1) = LIST_VALUE(t,j);
        LIST_MOVE(t,j)   = tmove;
        LIST_VALUE(t,j)  = tvalue;
      }
    }
  } 
}


/*
 * Stop the main injector loop
 * (not ponder)
 */ 
void Stop()
{
  brainStatus = PAUSED;
  watch.Reset();
  isSearchMove = false;
}

/*
 * Start the main injector loop
 * 
 */ 
void Start()
{
  watch.Start();
  // Must be done before each Search (this the corresp. function)
  //TT.Clear();
  nbProcessedNodes = 0;
  //sumNodeCutOffIndex = 0;
  if (!isSearchMove)
  {
    GetLegalMoves(currentPosition, moves);
  } // else moves are already provided
  // TODO: modify this part when implementing ponder mode
  // do this any time a new search is performed
  InitLegalMoves(moves);

  if(debugMode)
    DisplayMoves(moves);

  // reajust thinking time
  if (brainMode == ADJUSTTIME_MODE)
    UpdateThinkDelay();

  //std::cout <<  "Switch to THINKING status" << std::endl;
  brainStatus = THINKING;
}

void DisplayMoves(list_t& moves)
{
  Position tPos(currentPosition);
  tPos.board.DebugDisplay();
  ostringstream sb;
  sb << "info #### MOVES TO EVALUATE (" << moves.size << ")";
  int maxPerLine = 10;
  for(int i=0;i<moves.size;i++)
  {
    if((i % maxPerLine) == 0) sb << std::endl;
    Move m = moves.move[i];
    sb << ' ' << move_to_string(m) ;
  }
  std::cout <<  sb.str() << std::endl;
}

void SetDebug(bool isDebugOn)
{
  debugMode = isDebugOn;
}

void SetOption(string name, string value)
{
  
  if(name == "Hash")
  {
    int hashSize = atoi(value.c_str());
    TTable.ResizeMB(hashSize);  
  }
  /*
  else if(name == "UCI_LimitStrength")
  {
    UCI_LimitStrength = (value == "true");
  } 
  else if(name == "UCI_Elo")
  {
    UCI_Elo = atoi(value.c_str());
  }
   */
  else if(name == "UCI_AnalyseMode")
  {
    UCI_AnalyseMode = (value == "true");
  }
  else if(name == "Queen weight")
  {
    queenWeight = atoi(value.c_str());
  }
  else if(name == "Rook weight")
  {
    rookWeight = atoi(value.c_str());
  }
  else if(name == "Bishop weight")
  {
    bishopWeight = atoi(value.c_str());
  }
  else if(name == "Knight weight")
  {
    knightWeight = atoi(value.c_str());
  }
  else if(name == "Pawn weight")
  {
    pawnWeight = atoi(value.c_str());
  }
  
  else if(name == "King mobility")
  {
    kingMobilityFactor = atoi(value.c_str());
  }
  else if(name == "Queen mobility")
  {
    queenMobilityFactor = atoi(value.c_str());
  }
  else if(name == "Rook mobility")
  {
    rookMobilityFactor = atoi(value.c_str());
  }
  else if(name == "Bishop mobility")
  {
    bishopMobilityFactor = atoi(value.c_str());
  }
  else if(name == "Knight mobility")
  {
    knightMobilityFactor = atoi(value.c_str());
  }
  else if(name == "Pawn mobility")
  {
    pawnMobilityFactor = atoi(value.c_str());
  }
  else if(name == "Pawn structure")
  {
    pawnStructureFactor = atoi(value.c_str());
  }
  else if(name == "Smart Timer")
  {
    isSmartTiming = (value == "true");
  }
  else if(name == "Blackbox")
  {
    isBlackbox = (value == "true");
  }
  else 
  {
    std::cout <<  "Unkown option: " << name << std::endl;
    return;
  }
  // at this point the option is correctly set
  if(debugMode) std::cout <<  "option [" << name <<"] = " << value << std::endl;
}


// TESTING ONLY !!! (bad perfs)
bool IsCheckmate(Position& position)
{
  // new bitboard implementation
  list_t lm;
  GetLegalMoves(position, lm);
  return position.IsCheck() && LIST_IS_EMPTY(lm);
}


Move GetBestMove()
{
  return bestMove;
}


void SetDepthLevel(int depth)
{
  depthLevel = depth;
}

void SetTimeLevel(int timeLevel)
{
  brainMode = MOVETIME_MODE;
  thinkDelay = timeLevel;
}

void SetStartPosition()
{
  SetPosition("rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR w KQkq - 0 1");
}

void SetPosition(string strPos)
{
  //assert(brainStatus != THINKING);
  TTable.clear();
  clear_history();
  // Threads fuse
  while(brainStatus == THINKING)
  {
    Sleep(10);
  }
  currentPosition.SetFromFEN(strPos);
}

void Ponderhit()
{
  ////throw new NotImplementedException();
}

void SetSearchmovesMode(list_t & selectedMoves)
{
  // don't forget to switch to false once used
  moves = selectedMoves;
  isSearchMove = true;
}

void SetPonderMode()
{
  ////throw new NotImplementedException();
}

void SetWtime(int timeLeft)
{
  wtime = timeLeft;
  brainMode = ADJUSTTIME_MODE;
}

void SetBtime(int timeLeft)
{
  btime = timeLeft;
  brainMode = ADJUSTTIME_MODE;
}

void SetWinc(int increment)
{
  winc = increment;
  brainMode = ADJUSTTIME_MODE;
}

void SetBinc(int increment)
{
  binc = increment;
  brainMode = ADJUSTTIME_MODE;
}

void SetMovestogo(int movesUntilNextTimeControl)
{
  // TODO: use this value for time control init
  //	1-decrement for each move to zero
  //	2-make a time control
  //	3-reinit value and goto (1)	
  movestogo = movesUntilNextTimeControl;
	
}

void SetDepthMode(int pliesDepth)
{
  brainMode = DEPTH_MODE;
  depthLevel = pliesDepth; 
}

void SetNodesMode(int nodesDepth)
{
  brainMode = NODES_MODE;
  nodeMaxNumber = nodesDepth;
}

void SetMateMode(int mateDepth)
{
  brainMode = MATE_MODE;
  depthLevel = mateDepth;
}

void SetMovetimeMode(int searchTime)
{
  thinkDelay = searchTime;
  brainMode = MOVETIME_MODE;
}

void SetInfiniteMode()
{
  brainMode = INFINITE_MODE;
}

void SetVerboseMode(bool isVerboseMode)
{
  isVerbose = isVerboseMode;
}

void UpdateThinkDelay()
{
  int playerInc = 0;
  int playerTime = 0;
  //int oppInc = 0; // NOT USED FOR NOW
  int oppTime = 0;
  if (GetTurn() == WHITE)
    {
      playerInc = winc;
      playerTime = wtime;
      //oppInc = binc;
      oppTime = btime;
    } else {
    playerInc = binc;
    playerTime = btime;
    //oppInc = winc;
    oppTime = wtime;
  }
  int delay;
  int extraDelay;
  // Smart timing : the engine adjust its think time depending on 
  // the time left for itself and the opponent
  // TODO: improve calculation
  // TODO: link calculation to current nodeStack/score state
  if (isSmartTiming)
    {
      if (playerInc == 0)
	{
	  // TODO: improve time calculation when there is no inc

	  // no increment
	  // time must be managed carefully

      if (playerTime > oppTime)
		  {
        delay = (int)max(playerInc, playerTime / 20) + (playerTime-oppTime)/2;
		  }
	    else
		  {
		    delay = (int)max(playerInc, playerTime / 20) ;
		  }
	  }
    else
	  {
	    // base time
	    delay = playerInc;

	    // extra time taken from the clock
	    if(playerTime<10000){
	      // for timer < 10s try to save some time
	      extraDelay = playerTime*60/100;
	    }
	    else{
	      extraDelay = max(playerTime/min(40,max(5,abs(40-currentPosition.GetPlayedMoveNumber()))),8000);
	    }
	    delay += extraDelay;
	  }

    } 
  else
    // NO SMART TIMING
    {
      delay = (int)max(playerInc, playerTime / 20);
    }
  thinkDelay = delay;
  std::cout <<  "Think delay updated : " << thinkDelay << std::endl;
}


void DisplayBestPv(unsigned int maxDepth, uint64_t nodeNb, uint64_t time, Move* pv)
{

  uint64_t nps = (time == 0) ? 0 : ((1000*nodeNb) / time);
  unsigned int hashfull = TTable.GetUsagePerMil();
  string sScore = ScoreToString(bestScore);

  std::cout <<  "info"
  << " nodes "    << nodeNb 
  << " nps "      << nps
  << " depth "    << maxDepth
  << " seldepth " << selectiveDepth
  << " score "    << sScore
  << " time "     << time
  << " hashfull " << hashfull
  << " pv ";
  DisplayPv(pv,maxDepth);

  //printf("\ninfo string transposition table hits %d ", TTable.GetHits());

  std::cout <<  std::endl;
  if(debugMode) display_engine_cuttoff_accuracy(maxDepth);
  if(debugMode) display_engine_hashtable_usage();
}

void DisplayPv(list_t pv){
  for(int i=0;i<pv.size;i++){
    std::cout <<  move_to_string(pv.move[i]) << ' ';
  }
}

void DisplayPv(Move* pv, int maxDepth)
{
  for(int i=0; i<maxDepth;i++)
  {
    Move m = pv[i];
    if(m == MOVE_NULL)
      break;
    std::cout <<  move_to_string(m) << ' ';
  }
}

string ScoreToString(int score)
{
  char scoreStr[16];
  if( score < -VALUE_KNOWN_WIN || score > VALUE_KNOWN_WIN )
  {
    int plies = 0;
    if (score > 0)
	  {
	    plies = (VALUE_MATE - score);
    }
    else
    {
      plies = (-VALUE_MATE - score);
    }
    //mateIn = mateIn + plies;
    sprintf(scoreStr,"mate %d", plies);
  }
  else
  {
    sprintf(scoreStr,"cp %d", score);
  }
  return scoreStr;
}

Position& GetCurrentPosition()
{
  return currentPosition;
}

int GetBestScore()
{
  return bestScore;
}

string GetEngineInfo(){
  stringstream infos;
  infos << "id name OliveChess " << OLIVECHESS_VERSION << std::endl;
  infos << "id author Olivier HERAU" << std::endl;
  
  if(debugMode)
  {
    infos << "info OPTIMIZATION HasPopCnt=" << HasPopCnt << std::endl;
    infos << "info OPTIMIZATION Is64Bit  =" << Is64Bit << std::endl;
  }
  
  return infos.str();
}

string GetEngineOptions()
{	
  // TODO: simplify this method with static content
  //	EngineOptions.h should be removed as it is only used by the GUI

  stringstream sso;
  
  // OPTIONS TO IMPLEMENT FIRST
  sso << "option name Smart Timing type check default false" << std::endl;
  sso << "option name Hash type spin min 1 max 2048 default 16" << std::endl;
  //sso << "option name Style type combo default Standard var Dumb var Coward var Standard var Risky var Killer" << std::endl;
  sso << "option name UCI_Elo type spin min " << MIN_ENGINE_ELO << " max " << MAX_ENGINE_ELO << " default " << MAX_ENGINE_ELO << std::endl;
  sso << "option name UCI_LimitStrength type check default false" << std::endl;
  sso << "option name UCI_AnalyseMode type check default false" << std::endl;
    
  // Chessmen values
  sso << "option name Queen weight type spin min 0 max 2000 default 900" << std::endl;
  sso << "option name Rook weight type spin min 0 max 2000 default 500" << std::endl;
  sso << "option name Bishop weight type spin min 0 max 2000 default 325" << std::endl;
  sso << "option name Knight weight type spin min 0 max 2000 default 325" << std::endl;
  sso << "option name Pawn weight type spin min 0 max 2000 default 85" << std::endl;
  
  // Mobility factor
  sso << "option name King mobility type spin min 0 max 100 default 100" << std::endl;
  sso << "option name Queen mobility type spin min 0 max 100 default 100" << std::endl;
  sso << "option name Rook mobility type spin min 0 max 100 default 100" << std::endl;
  sso << "option name Bishop mobility type spin min 0 max 100 default 100" << std::endl;
  sso << "option name Knight mobility type spin min 0 max 100 default 100" << std::endl;
  sso << "option name Pawn mobility type spin min 0 max 100 default 100" << std::endl;

  // Pawn structure
  sso << "option name Pawn structure type spin min 0 max 20 default 5" << std::endl;
  
  //sso << "option name Selectivity type spin default 2 min 0 max 4" << std::endl;

  /* STANDARD GENERIC OPTIONS
  sso << "option name NalimovCache type spin min 1 max 256 default 32" << std::endl;
  sso << "option name NalimovPath type string default " << std::endl;
  sso << "option name Ponder type check default true" << std::endl;
  sso << "option name OwnBook type check default true" << std::endl;
  sso << "option name MultiPV type spin min 1 max 127 default 1" << std::endl;
  sso << "option name UCI_SearchMoves type check default true" << std::endl;
  sso << "option name UCI_ShowCurrLine type check default false" << std::endl;
  sso << "option name UCI_Chess960 type check default false" << std::endl;
  sso << "option name UCI_Opponent type string default " << std::endl;
  */
  /* HIARCS options
  option name CPU Usage type spin min 1 max 100 default 100
  option name OwnBook type check default true
  option name Book Learning type check default false
  option name Book Mode type combo default Tournament var Wild var Surprise var Dynamic var Tournament
  option name Book Positions type combo default 2581613 var 2581613
  option name Book Status type combo default Hiarcs13cBook.hcs OK var Hiarcs13cBook.hcs OK
  option name Use Tablebases type combo var Often var Normally var Rarely var Never default Never
  option name Swindle Opponent type check default true
  option name Smart Search type check default true
  option name Optimistic Search type check default true
  option name Combinations type check default true
  option name Search Selectivity type spin min 0 max 7 default 4
  option name Playing Style type combo default Active var Active var Aggressive var Solid
  */  

  return sso.str();
}

void SetSmartTiming(bool value)
{
	isSmartTiming = value;
}


Color GetTurn(){
	return currentPosition.gameState.turn;
}


void display_engine_monitor()
{
  long time = watch.GetElapsedTime();
  uint64_t nps = (time == 0) ? 0 : ((1000*nbProcessedNodes) / time);
  std::cout <<  "info"
    << " depth " << depthLevel
    << " time " << time 
    << " nodes " << nbProcessedNodes 
    << " nps " << nps 
    << " tbhits " << TTable.GetHits() << std::endl;
  std::cout << "info hashfull " << TTable.GetUsagePerMil() << std::endl;
}


void display_engine_hashtable_usage()
{
  std::cout << "info hashfull " << TTable.GetUsagePerMil() << std::endl;
  std::cout << "info tbhits " << TTable.GetHits() << std::endl;
}

void display_engine_cuttoff_accuracy(int maxDepth)
{
  int cuttoffaccuracy = (nbCurrSearchTotalGenNodes == 0) ? 0 : 100.0-(100.0*nbCurrSearchProcessedNodes / nbCurrSearchTotalGenNodes);
  std::cout << "info string cut-off accuracy: " << cuttoffaccuracy <<  '%' << std::endl;
  if(maxDepth<20)
  {
    cutOffStatTable[maxDepth] = cuttoffaccuracy;
  }
}

void display_engine_movelist()
{
  for(int i=0;i<moves.size;i++)
  {
    std::cout <<  "info currmove " << move_to_string(moves.move[i]) << " currmovenumber " << (i+1) << std::endl;
  }
}

void display_engine_depthstats(int* statTable,int maxDepth)
{
  int tableSize = 20;
  int minIndex=2;
  int maxIndex=min(maxDepth,tableSize);
  int range = maxIndex-minIndex+1 ;
  char bar = '8';
  //char halfbar = 'o';
  char nobar = ' ';
  //int minVal=0;
  int maxVal=100;
  int i;
  if(maxIndex<minIndex) return;
  /*for(int i=0;i<tableSize;i++)
  {
    minVal = min(minVal,statTable[i]);
    maxVal = max(maxVal,statTable[i]);
  }*/
  
  cout << "100%" << endl;
  int step = 10;
  for(int v=100;v>=0;v-=step)
  {
    cout << " | ";
    for(i=minIndex;i<=maxIndex;i++)
    {
      if( ((statTable[i] * 100.0)/maxVal) > v)
        cout << bar << bar << nobar;
      //else if( ((statTable[i] * 100.0)/maxVal) > (v-(step/2)))
      //  cout << halfbar << halfbar << nobar; 
      else
        cout << nobar << nobar << nobar;  
    }
    cout << endl;
  }
  cout << " +-";
  for(i=minIndex;i<=maxIndex;i++)
  {
    cout << "---";
  }
  cout << endl << "0% ";
  for(i=minIndex;i<=maxIndex;i++)
  {
    if(i<10)
      cout << ' ';
    cout << i << ' ';
  }
  cout << endl;
  
  int sum=0;
  for(i=minIndex;i<=maxIndex;i++)
  {
    sum += statTable[i];
  }
  int mean = (sum/range);
  
  sum = 0;
  for(i=minIndex;i<=maxIndex;i++)
  {
    sum += (statTable[i] - mean)*(statTable[i] - mean);
  }
  int stdDev = sqrt((float)(sum/range));
  cout << "mean=" << mean << " stdDev=" << stdDev << endl; 
}
