#include <cassert>
#include <string>
#include "Move.h"
#include "Position.h"

using std::string;

const string move_to_string(Move m) {

  Square from = from_sq(m);
  Square to = to_sq(m);
  string promotion;
  
  if (m == MOVE_NONE) return "(none)";
  if (m == MOVE_NULL) return "0000";

  if (is_promotion(m))
      promotion = char(tolower(piece_type_to_char(promotion_type(m))));

  return square_to_string(from) + square_to_string(to) + promotion;
}

// should return MOVE_NONE if unlegal
Move string_to_move(Position& pos, const string& str) {
	
  Square from = make_square((File)(str[0]-'a'),(Rank)(str[1]-'1'));
  Square to   = make_square((File)(str[2]-'a'),(Rank)(str[3]-'1'));
  
  Piece p = pos.board.pieces[from];

  if((p == W_KING) || (p == B_KING))
  {
    if((str[0] == 'e') && (str[2] == 'g' || str[2] == 'c'))
      return make_castle(from,to);
  }
  else if((p == W_PAWN) || (p == B_PAWN))
  {
    if((str[0] != str[2]) && pos.IsEmpty(to))
      return make_enpassant(from,to);
  }
	if(str.size() == 5){
		switch(str[4])
		{
		  case 'q':
			  return make_promotion(from,to,QUEEN);
		  case 'r':
			  return make_promotion(from,to,ROOK);
		  case 'b':
			  return make_promotion(from,to,BISHOP);
		  case 'n':
			  return make_promotion(from,to,KNIGHT);
		}
	}
	
  return make_move(from,to);
}

void print(list_t& moveList)
{
  for(int i=0;i<moveList.size;i++)
  {
    std::cout <<  " " << move_to_string(moveList.move[i]);
  }
  

}

