#include "MoveGen.h"
#include "Bitboard.h"
#include "Bitboard.h"
#include "Attack.h"
#include "Engine.h"


/*
 * Initialize moves at level zero before 
 * the search begins (only the move is kept)
 */
void InitLegalMoves(list_t & legalMoves)
{
  for(int i=0;i<LIST_SIZE(legalMoves);i++)
  {
    LIST_VALUE(legalMoves,i)  = VALUE_ZERO;
  }
}



// TODO: refactor entirely the move generation to accurately generate the different types of moves

/* FROM FRUIT
//move_gen.h
extern void gen_legal_moves (list_t * list, board_t * board);

extern void gen_moves       (list_t * list, const board_t * board);
extern void gen_captures    (list_t * list, const board_t * board);
extern void gen_quiet_moves (list_t * list, const board_t * board);

extern void add_pawn_move   (list_t * list, int from, int to);
extern void add_promote     (list_t * list, int move);

//move_gen.cpp
static void add_moves               (list_t * list, const board_t * board);
static void add_captures            (list_t * list, const board_t * board);
static void add_quiet_moves         (list_t * list, const board_t * board);

static void add_promotes            (list_t * list, const board_t * board);
static void add_en_passant_captures (list_t * list, const board_t * board);
static void add_castle_moves        (list_t * list, const board_t * board);


END FRUIT CODE */ 


inline bool IsFiftyMovesDraw(Position& pos) 
{
  return (pos.gameState.halfMoveClock >= 50);
}


inline bool contains_promotion(list_t& _moves)
{
  for(int i=0;i<_moves.size;i++)
  {
    if(is_promotion(_moves.move[i]) ) 
      return true;
  }

  return false; 
}

inline bool contains_move(list_t& _moves, Move m)
{
  for(int i=0;i<_moves.size;i++)
  {
    if(_moves.move[i] == m ) 
      return true;
  }

  return false; 
}

/**
* Get legal moves - Main bottleneck
* @param currentPosition
* @return
*/
int GetLegalMoves(Position & pos, list_t & legalMoves)
{
  LIST_CLEAR(legalMoves);
  if (IsFiftyMovesDraw(pos))
    return 0;

  // 1- retrieve all moves
  GetPossibleMoves(pos, legalMoves);

  // 2 - prune illegal moves
  // copy elements and drop the rest of the list
  // no temporary vector 
  Color turn = pos.GetTurn();
  int lastLegalMoveIndex = 0;
  for(int i=0 ; i < legalMoves.size ; i++ )
  {
    Move m = legalMoves.move[i];
    pos.Play(m);
    if(!pos.IsCheck(turn))
    {
      if(i != lastLegalMoveIndex) // do the copy only if needed
        legalMoves.move[lastLegalMoveIndex] = m;
      //// update score with heuristic history value in order to sort
      //legalMoves.score[lastLegalMoveIndex] = history[BLACK-pos.GetTurn()][from_sq(m)][to_sq(m)];
      lastLegalMoveIndex++;		// the move is legal, always increment
    }
    pos.Unplay(m);
  }
  // resize the list
  legalMoves.size = lastLegalMoveIndex;

  // 3- Sort moves by potential score
  //SetHistoryValue(legalMoves,pos);
  //SetPotentialValue(legalMoves);
  SetMvvLvaValue(legalMoves,pos);
  Sort(legalMoves);
  //pv_reorder(legalMoves,pvhash); // push pv moves first
  //TTable.PushUpKillerMoves(legalMoves);
  
  
  return legalMoves.size;
}


void SetPotentialValue(list_t& list)
{  
  for (int i = 0; i < list.size; i++) 
  {
    Move move = LIST_MOVE(list,i);
    LIST_VALUE(list,i) = -potentialValue(move);
  }
}

// MVV-LVA - Most Valuable Victim - Least Valuable Aggressor
void SetMvvLvaValue(list_t& list, Position& p)
{  
  for (int i = 0; i < list.size; i++) 
  {
    Move move = LIST_MOVE(list,i);
    LIST_VALUE(list,i) = MvvLvaValue(p,move);
  }
}


void SetHistoryValue(list_t& list, Position& p)
{
  for (int i = 0; i < list.size; i++) 
  {
    Move m = LIST_MOVE(list,i);
    LIST_VALUE(list,i) = -history[p.GetPiece(from_sq(m))][to_sq(m)];
  }
}

// MVV-LVA - Most Valuable Victim - Least Valuable Aggressor
int MvvLvaValue(Position& pos, Move move)
{
  // best captures first
  int captureEval = 2000*GetVal(pos.GetPiece(to_sq(move)));
  // lowest piece moved for capture first
  int pieceEval = 2000-GetVal(pos.GetPiece(from_sq(move)));
  return captureEval + pieceEval;
}


int potentialValue(Move move) 
{  
  return ((move & 07777) << 2) | ((move >> 12) & 3);
}

//Piece GetCPiece(Position& pos, Move move)
//{
//  return pos.board.pieces[to_sq(move)];
//}
//
//Piece GetMPiece(Position& pos, Move move)
//{
//  return pos.board.pieces[from_sq(move)];
//}

int GetVal(Piece p)
{
  switch(p)
  {
  case W_KING:
  case B_KING:
    return 0;
    break;
  case W_QUEEN:
  case B_QUEEN:
    return queenWeight;
    break;
  case W_ROOK:
  case B_ROOK:
    return rookWeight;
    break;
  case W_BISHOP:
  case B_BISHOP:
    return bishopWeight;
    break;
  case W_KNIGHT:
  case B_KNIGHT:
    return knightWeight;
    break;
  case W_PAWN:
  case B_PAWN:
    return pawnWeight;
    break;
  default:
    break;
  }
  return 0;
}



// Give the list of possible moves
void GetPossibleMoves(Position& position, list_t & moves)
{
  //LIST_CLEAR(moves);
  for (Square square = SQ_A1; square <= SQ_H8; square++)
  {
    if (!position.IsEmptyOrOpponent(square))
    {
      GetPossibleMoves(position, square, moves);
    }
  }
}


/*
// Give the list of possible moves
// uses Bitboards instead of parsing the board
void GetPossibleMovesBB(Position& position, list_t & moves)
{
  GetPossibleKingMoves(position, moves, moves, moves);
  GetPossibleQueenMoves(position, moves, moves);
  GetPossibleRookMoves(position, moves, moves);
  GetPossibleKnightMoves(position, moves, moves);
  GetPossibleBishopMoves(position, moves, moves);
  GetPossiblePawnMoves(position, moves, moves, moves, moves);
}
*/

/*
void GetPossibleKingMoves(Position & p, list_t & cMoves, list_t & qMoves, list_t& castleMoves )
{
  Color side = p.GetTurn();
  Color opp = BLACK-side;
  int c = (side == WHITE) ? 0 : 8;
  
  Bitboard fbb = p.board.piecesBB[c+KING];
  Square fsq = pop_1st_bit(&fbb);
  
  Bitboard tbb = p.board.emptyBB & PseudoAttacks[KING][fsq];
  while(tbb)
  {
    Square tsq = pop_1st_bit(&tbb);
    LIST_ADD(qMoves,make_move(fsq,tsq));
  }
  tbb = p.board.colorBB[opp] & PseudoAttacks[KING][fsq];
  while(tbb)
  {
    Square tsq = pop_1st_bit(&tbb);
    LIST_ADD(cMoves,make_move(fsq,tsq));
  }
  
  if (p.GetTurn()==WHITE)
  {
    if (p.gameState.whiteCanCastleQueenSide)
      if(p.IsEmpty(SQ_B1))
        if( p.IsEmpty(SQ_C1))
          if(p.IsEmpty(SQ_D1))
            if(!p.IsCheck(WHITE, SQ_E1))
              if(!p.IsCheck(WHITE, SQ_D1))
                if(!p.IsCheck(WHITE, SQ_C1))
                  LIST_ADD(castleMoves,make_castle(SQ_E1, SQ_C1));
    if (p.gameState.whiteCanCastleKingSide)
      if(p.IsEmpty(SQ_E1))
        if(p.IsEmpty(SQ_F1))
          if(p.IsEmpty(SQ_G1))
            if(!p.IsCheck(WHITE, SQ_E1))
              if(!p.IsCheck(WHITE, SQ_F1))
                if(!p.IsCheck(WHITE, SQ_G1))
                  LIST_ADD(castleMoves,make_castle(SQ_E1, SQ_G1));
  }
  else
  {
    if (p.gameState.blackCanCastleQueenSide)
      if(p.IsEmpty(SQ_B8))
        if( p.IsEmpty(SQ_C8))
          if(p.IsEmpty(SQ_D8))
            if(!p.IsCheck(BLACK, SQ_E8))
              if(!p.IsCheck(BLACK, SQ_D8))
                if(!p.IsCheck(BLACK, SQ_C8))
                  LIST_ADD(castleMoves,make_castle(SQ_E8, SQ_C8));
    if (p.gameState.blackCanCastleKingSide)
      if(p.IsEmpty(SQ_E8))
        if(p.IsEmpty(SQ_F8))
          if(p.IsEmpty(SQ_G8))
            if(!p.IsCheck(BLACK, SQ_E8))
              if(!p.IsCheck(BLACK, SQ_F8))
                if(!p.IsCheck(BLACK, SQ_G8))
                  LIST_ADD(castleMoves,make_castle(SQ_E8, SQ_G8));
  }
}
*/

/*
void GetPossibleKnightMoves(Position & p, list_t & cMoves, list_t & qMoves )
{
  Color side = p.GetTurn();
  Color opp = BLACK-side;
  int c = (side == WHITE) ? 0 : 8;
  
  Bitboard fbb = p.board.piecesBB[c+KNIGHT];
  Square fsq = pop_1st_bit(&fbb);
  
  Bitboard tbb = p.board.emptyBB & PseudoAttacks[KNIGHT][fsq];
  while(tbb)
  {
    Square tsq = pop_1st_bit(&tbb);
    LIST_ADD(qMoves,make_move(fsq,tsq));
  }
  tbb = p.board.colorBB[opp] & PseudoAttacks[KNIGHT][fsq];
  while(tbb)
  {
    Square tsq = pop_1st_bit(&tbb);
    LIST_ADD(cMoves,make_move(fsq,tsq));
  }
  
}


void GetPossibleQueenMoves(Position & p, list_t & cMoves, list_t & qMoves )
{
  GetPossibleRookMoves(p, cMoves, qMoves);
  GetPossibleBishopMoves(p, cMoves, qMoves);
  
}
void GetPossibleRookMoves(Position & p, list_t & cMoves, list_t & qMoves )
{
}

void GetPossibleBishopMoves(Position & p, list_t & cMoves, list_t & qMoves )
{
}
void GetPossiblePawnMoves(Position & p, list_t & cMoves, list_t & qMoves, list_t& promMoves, list_t& enpassantMoves )
{
}
*/

/**
* Get the possible moves for one piece
* 
* @param currentPosition
* @param square
* @return
*/
void GetPossibleMoves(Position& position, Square square, list_t & possibleMoves)
{
  Piece pieceCode = position.GetPiece(square);

  if ((position.IsBlackToMove() && position.IsPieceBlack(pieceCode))
    || (position.IsWhiteToMove() && position.IsPieceWhite(pieceCode)))
  {
    switch (pieceCode)
    {
    case W_PAWN:
    case B_PAWN:
      GetPossiblePawnMoves(position, square, possibleMoves);
      break;
    case W_BISHOP:
    case B_BISHOP:
      GetPossibleBishopMoves(position, square, possibleMoves);
      break;
    case W_KNIGHT:
    case B_KNIGHT:
      GetPossibleKnightMoves(position, square, possibleMoves);
      break;
    case W_ROOK:
    case B_ROOK:
      GetPossibleRookMoves(position, square, possibleMoves);
      break;
    case W_QUEEN:
    case B_QUEEN:
      GetPossibleQueenMoves(position, square, possibleMoves);
      break;
    case W_KING:
    case B_KING:
      GetPossibleKingMoves(position, square, possibleMoves);
      break;
    default:
      break;
    }
  }
}


/**
* Get possible moves for King
* 
* @param currentPosition
* @param square
* @return
*/
void GetPossibleKingMoves(Position& position, Square square, list_t & possibleKingMoves)
{
  File col = file_of(square);
  Rank row = rank_of(square);

  if (col < FILE_H && position.IsEmptyOrOpponent(square+DELTA_E))
  {
    LIST_ADD(possibleKingMoves,make_move(square, square+DELTA_E));
  }
  if (col < FILE_H && row < RANK_8 && position.IsEmptyOrOpponent(square+DELTA_NE))
  {
    LIST_ADD(possibleKingMoves,make_move(square, square+DELTA_NE));
  }
  if (row < RANK_8 && position.IsEmptyOrOpponent(square+DELTA_N))
  {
    LIST_ADD(possibleKingMoves,make_move(square, square+DELTA_N));
  }
  if (col > FILE_A && row < RANK_8 && position.IsEmptyOrOpponent(square+DELTA_NW))
  {
    LIST_ADD(possibleKingMoves,make_move(square, square+DELTA_NW));
  }
  if (col > FILE_A && position.IsEmptyOrOpponent(square+DELTA_W))
  {
    LIST_ADD(possibleKingMoves,make_move(square, square+DELTA_W));
  }
  if (col > FILE_A && row > RANK_1 && position.IsEmptyOrOpponent(square+DELTA_SW))
  {
    LIST_ADD(possibleKingMoves,make_move(square, square+DELTA_SW));
  }
  if (row > RANK_1 && position.IsEmptyOrOpponent(square+DELTA_S))
  {
    LIST_ADD(possibleKingMoves,make_move(square, square+DELTA_S));
  }
  if (col < FILE_H && row > RANK_1 && position.IsEmptyOrOpponent(square+DELTA_SE))
  {
    LIST_ADD(possibleKingMoves,make_move(square, square+DELTA_SE));
  }

  if (position.GetTurn()==WHITE)
  {
    if (position.gameState.whiteCanCastleQueenSide)
      if(position.IsEmpty(SQ_B1))
        if( position.IsEmpty(SQ_C1))
          if(position.IsEmpty(SQ_D1))
            if(!position.IsCheck(WHITE, SQ_E1))
              if(!position.IsCheck(WHITE, SQ_D1))
                if(!position.IsCheck(WHITE, SQ_C1))
                  LIST_ADD(possibleKingMoves,make_castle(SQ_E1, SQ_C1));
    if (position.gameState.whiteCanCastleKingSide)
      if(position.IsEmpty(SQ_E1))
        if(position.IsEmpty(SQ_F1))
          if(position.IsEmpty(SQ_G1))
            if(!position.IsCheck(WHITE, SQ_E1))
              if(!position.IsCheck(WHITE, SQ_F1))
                if(!position.IsCheck(WHITE, SQ_G1))
                  LIST_ADD(possibleKingMoves,make_castle(SQ_E1, SQ_G1));
  }
  else
  {
    if (position.gameState.blackCanCastleQueenSide)
      if(position.IsEmpty(SQ_B8))
        if( position.IsEmpty(SQ_C8))
          if(position.IsEmpty(SQ_D8))
            if(!position.IsCheck(BLACK, SQ_E8))
              if(!position.IsCheck(BLACK, SQ_D8))
                if(!position.IsCheck(BLACK, SQ_C8))
                  LIST_ADD(possibleKingMoves,make_castle(SQ_E8, SQ_C8));
    if (position.gameState.blackCanCastleKingSide)
      if(position.IsEmpty(SQ_E8))
        if(position.IsEmpty(SQ_F8))
          if(position.IsEmpty(SQ_G8))
            if(!position.IsCheck(BLACK, SQ_E8))
              if(!position.IsCheck(BLACK, SQ_F8))
                if(!position.IsCheck(BLACK, SQ_G8))
                  LIST_ADD(possibleKingMoves,make_castle(SQ_E8, SQ_G8));
  }
}


/**
* Get possible moves for Queen
* 
* @param currentPosition
* @param square
* @return
*/
void GetPossibleQueenMoves(Position& position, Square square, list_t & possibleQueenMoves)
{
  GetPossibleRookMoves(position, square, possibleQueenMoves);
  GetPossibleBishopMoves(position, square, possibleQueenMoves);
}

/*
* Get possible moves for Rook
*/
void GetPossibleRookMoves(Position& position, Square square, list_t & possibleRookMoves)
{
  //File col = file_of(square); //NOT USED
  Rank row = rank_of(square);

  // Horizontal moves right
  Square sqStart = square + DELTA_E;
  Square sqEnd = (Square)(8 * (row+1));
  for (Square sq = sqStart; sq < sqEnd; sq++)
  {
    if (position.IsEmptyOrOpponent(sq))
    {
      LIST_ADD(possibleRookMoves,make_move(square, sq));
      if (!position.IsEmpty(sq))
        break;
    }
    else
      break;
  }

  // Horizontal moves left
  sqStart = square + DELTA_W;
  sqEnd = (Square)(8 * row - 1);
  for (Square sq = sqStart; sq > sqEnd; sq--)
  {
    if (position.IsEmptyOrOpponent(sq))
    {
      LIST_ADD(possibleRookMoves,make_move(square, sq));
      if (!position.IsEmpty(sq))
        break;
    }
    else
      break;
  }


  // Vertical moves up
  sqStart = (Square)(square + 8);
  sqEnd = (Square)64;
  for (Square sq = sqStart; sq < sqEnd; sq+=DELTA_N)
  {
    if (position.IsEmptyOrOpponent(sq))
    {
      LIST_ADD(possibleRookMoves,make_move(square, sq));
      if (!position.IsEmpty(sq))
        break;
    }
    else
      break;
  }

  // Vertical moves down
  sqStart = square + DELTA_S;
  sqEnd = (Square)-1;
  for (Square sq = sqStart; sq > sqEnd; sq+=DELTA_S)
  {
    if (position.IsEmptyOrOpponent(sq))
    {
      LIST_ADD(possibleRookMoves,make_move(square, sq));
      if (!position.IsEmpty(sq))
        break;
    }
    else
      break;
  }
}


/**
* Get possible moves for Knight
* 
* @param currentPosition
* @param square
* @return
*/
void GetPossibleKnightMoves(Position& position, Square square, list_t & possibleKnightMoves)
{
  int col = file_of(square);
  int row = rank_of(square);

  if (col > FILE_A && row < RANK_7 && position.IsEmptyOrOpponent(square + DELTA_N + DELTA_NW))
  {
    LIST_ADD(possibleKnightMoves,make_move(square, square + DELTA_N + DELTA_NW));
  }
  if (col < FILE_H && row < RANK_7 && position.IsEmptyOrOpponent(square+ DELTA_N + DELTA_NE))
  {
    LIST_ADD(possibleKnightMoves,make_move(square, square + DELTA_N + DELTA_NE));
  }
  if (col < FILE_G && row < RANK_8 && position.IsEmptyOrOpponent(square+ DELTA_E + DELTA_NE))
  {
    LIST_ADD(possibleKnightMoves,make_move(square, square + DELTA_E + DELTA_NE));
  }
  if (col < FILE_G && row > RANK_1 && position.IsEmptyOrOpponent(square+ DELTA_E + DELTA_SE))
  {
    LIST_ADD(possibleKnightMoves,make_move(square, square + DELTA_E + DELTA_SE));
  }
  if (col < FILE_H && row > RANK_2 && position.IsEmptyOrOpponent(square+ DELTA_S + DELTA_SE))
  {
    LIST_ADD(possibleKnightMoves,make_move(square, square + DELTA_S + DELTA_SE));
  }
  if (col > FILE_A && row > RANK_2 && position.IsEmptyOrOpponent(square+ DELTA_S + DELTA_SW))
  {
    LIST_ADD(possibleKnightMoves,make_move(square, square + DELTA_S + DELTA_SW));
  }
  if (col > FILE_B && row > RANK_1  && position.IsEmptyOrOpponent(square+ DELTA_W + DELTA_SW))
  {
    LIST_ADD(possibleKnightMoves,make_move(square, square + DELTA_W + DELTA_SW));
  }
  if (col > FILE_B && row < RANK_8 && position.IsEmptyOrOpponent(square + DELTA_W + DELTA_NW))
  {
    LIST_ADD(possibleKnightMoves,make_move(square, square + DELTA_W + DELTA_NW));
  }

}


/**
* Get possible moves for Bishop
* 
* @param currentPosition
* @param square
* @return
*/
void GetPossibleBishopMoves(Position& position, Square square, list_t & possibleBishopMoves)
{
  File col = file_of(square);
  //Rank row = rank_of(square); // NOT USED

  // Up-right moves
  Square sqStart = square + DELTA_NE;
  Square sqEnd = (Square)min(64, square + 9 * (8 - col));
  for (Square sq = sqStart; sq < sqEnd; sq += DELTA_NE)
  {
    if (position.IsEmptyOrOpponent(sq))
    {
      LIST_ADD(possibleBishopMoves,make_move(square, sq));
      if (!position.IsEmpty(sq)) 
        break;
    }
    else 
      break;
  }

  // Down-right moves
  sqStart = square + DELTA_SE;
  sqEnd = (Square)max(-1, square - 7 * (8 - col));
  for (Square sq = sqStart; sq > sqEnd ; sq += DELTA_SE)
  {
    if (position.IsEmptyOrOpponent(sq))
    {
      LIST_ADD(possibleBishopMoves,make_move(square, sq));
      if (!position.IsEmpty(sq)) 
        break;
    }
    else 
      break;
  }

  // Down-left moves
  sqStart = square + DELTA_SW;
  sqEnd = (Square)max(-1, square - 9 * (col + 1));
  for (Square sq = sqStart ; sq > sqEnd; sq += DELTA_SW)
  {
    if (position.IsEmptyOrOpponent(sq))
    {
      LIST_ADD(possibleBishopMoves,make_move(square, sq));
      if (!position.IsEmpty(sq))
        break;
    }
    else 
      break;
  }

  // Up-left moves
  sqStart = square + DELTA_NW;
  sqEnd = (Square)min(64, square + 7 * (col + 1));
  for (Square sq = sqStart; sq < sqEnd; sq += DELTA_NW)
  {
    if (position.IsEmptyOrOpponent(sq))
    {
      LIST_ADD(possibleBishopMoves,make_move(square, sq));
      if (!position.IsEmpty(sq))
        break;
    }
    else
      break;
  }
}


/**
* Get possible moves for Pawn
* 
* @param currentPosition
* @param square
* @return
*/
void GetPossiblePawnMoves(Position& position, Square square, list_t & possiblePawnMoves)
{
  list_t pawnMoves;
  LIST_CLEAR(pawnMoves);
  Piece piece = position.GetPiece(square);
  File col = file_of(square);
  Rank row = rank_of(square);

  if (piece == W_PAWN && row < RANK_8)
  {
    // Pawn's move
    if (position.GetPiece(square+DELTA_N) == NO_PIECE)
    {
      LIST_ADD(pawnMoves,make_move(square, square + DELTA_N));
    }
    // pawn's double move
    if (row == RANK_2 && (position.GetPiece(square+DELTA_N) == NO_PIECE) 
      && (position.GetPiece(square+DELTA_NN) == NO_PIECE))
    {
      LIST_ADD(pawnMoves,make_move(square, square + DELTA_NN));
    }
    // Pawn capture
    if (col < FILE_H && (position.IsPieceBlack(position.GetPiece(square+DELTA_NE))))
    {
      LIST_ADD(pawnMoves,make_move(square, square + DELTA_NE));
    }
    // Pawn en-passant
    if (col < FILE_H && position.gameState.lastEnPassant == (square+DELTA_NE))
    {
      LIST_ADD(pawnMoves,make_enpassant(square, square + DELTA_NE));
    }   
    // capture
    if (col > FILE_A && position.IsPieceBlack(position.GetPiece(square+DELTA_NW)))
    {
      LIST_ADD(pawnMoves,make_move(square, square + DELTA_NW));
    }
    // Pawn en-passant
    if (col > FILE_A && position.gameState.lastEnPassant == (square+DELTA_NW))
    {
      LIST_ADD(pawnMoves,make_enpassant(square, square + DELTA_NW));
    }
  }
  if (piece == B_PAWN && row > RANK_1)
  {
    // Pawn's move
    if (position.GetPiece(square+DELTA_S) == NO_PIECE)
    {
      LIST_ADD(pawnMoves,make_move(square, square + DELTA_S));
    }
    // pawn's double move
    if (row == RANK_7 && (position.GetPiece(square + DELTA_S) == NO_PIECE) 
      && (position.GetPiece(square + DELTA_SS) == NO_PIECE))
    {
      LIST_ADD(pawnMoves,make_move(square, square + DELTA_SS));
    }
    // Pawn capture
    if (col < FILE_H && position.IsPieceWhite(position.GetPiece(square + DELTA_SE)))
    {
      LIST_ADD(pawnMoves,make_move(square, square + DELTA_SE));
    }
    // en passant
    if (col < FILE_H && position.gameState.lastEnPassant == (square + DELTA_SE))
    {
      LIST_ADD(pawnMoves,make_enpassant(square, square + DELTA_SE));
    }
    // capture
    if (col > FILE_A && position.IsPieceWhite(position.GetPiece(square + DELTA_SW)))
    {
      LIST_ADD(pawnMoves,make_move(square, square + DELTA_SW));
    }
    // en passant
    if (col > FILE_A && position.gameState.lastEnPassant == (square + DELTA_SW))
    {
      LIST_ADD(pawnMoves,make_enpassant(square, square + DELTA_SW));
    }
  }

  // adding alternate promotions
  for(int i=0;i<pawnMoves.size;i++)
  {
    Move m = pawnMoves.move[i];
    Rank r = rank_of(to_sq(m));
    if ((r == RANK_1) || (r == RANK_8))
    {
      LIST_ADD(possiblePawnMoves,make_promotion(from_sq(m), to_sq(m), QUEEN));
      LIST_ADD(possiblePawnMoves,make_promotion(from_sq(m), to_sq(m), ROOK));
      LIST_ADD(possiblePawnMoves,make_promotion(from_sq(m), to_sq(m), BISHOP));
      LIST_ADD(possiblePawnMoves,make_promotion(from_sq(m), to_sq(m), KNIGHT));
    }
    else
    {
      LIST_ADD(possiblePawnMoves,m);
    }
  }
}

/*
// new faster version - different name to force compile errors on the switch
int GenLegalMoves(Position& pos, list_t& moves)
{
  LIST_CLEAR(moves);
  if (IsFiftyMovesDraw(pos)) return 0;

  attack_t attackSide, attackOpp;

  Color side = pos.GetTurn();
  Color opp = (pos.GetTurn()==WHITE) ? BLACK : WHITE;

  get_attacks(pos.board, side, attackSide);
  get_attacks(pos.board, opp, attackOpp); // can be done only once
  Bitboard pinned = get_pinned(pos.board, side);

  std::cout <<  "Pinned" << std::endl;
  print(pinned);


  if(IsInCheck(pos.board, side, attackOpp))
  {
    GenEvasionMoves(pos, attackSide, attackOpp, pinned, moves);
  }
  else
  {
    GenCaptureMoves(pos, attackSide, attackOpp, pinned, moves);
    GenQuietMoves(pos, attackSide, attackOpp, pinned, moves);
    GenCastleMoves(pos, attackSide, attackOpp, pinned, moves);
  }

  return moves.size;
}
*/

/*
// 1.1 if in check, GetLegalEvade 
//     i.e. only moves removing check: 
//     -killing the attacking piece
//  or -moving king to a safe square
//  or -blocking sliding piece attack
//      (casteling is forbidden)
void GenEvasionMoves(Position& pos, attack_t& attackSide, attack_t& attackOpp, Bitboard& pinned, list_t& moves)
{
  //Color side = pos.GetTurn();
  //int s = (pos.GetTurn()==WHITE) ? 0 : 8;
  //int fw = (pos.GetTurn()==WHITE) ? 8 : -8;
  //Bitboard bbf, bbt;
  //int fsq,tsq;

  // 1-king escape to a free square
  GenQuietKingMoves(pos, attackSide, attackOpp, pinned, moves);

  // 2-block slider attack
  //   possible only if there is one attacker
  // TODO implement block slider attack


  // 3-remove attacker
  //   possible only if there is one attacker
  // TODO implement remove attacker



}
*/

/*
void GenQuietMoves(Position& pos, attack_t& attackSide, attack_t& attackOpp, Bitboard& pinned, list_t& moves)
{

  GenQuietPawnMoves(pos, attackSide, attackOpp, pinned, moves);
  GenQuietKnightMoves(pos, attackSide, attackOpp, pinned, moves);
  GenQuietKingMoves(pos, attackSide, attackOpp, pinned, moves);

  // could be regrouped for better perfs ?
  GenQuietRookMoves(pos, attackSide, attackOpp, pinned, moves);
  GenQuietBishopMoves(pos, attackSide, attackOpp, pinned, moves);
  GenQuietQueenMoves(pos, attackSide, attackOpp, pinned, moves);

}


void GenCaptureMoves(Position& pos, attack_t& attackSide, attack_t& attackOpp, Bitboard& pinned, list_t& moves)
{

  GenCapturePawnMoves(pos, attackSide, attackOpp, pinned, moves);
  GenCaptureKnightMoves(pos, attackSide, attackOpp, pinned, moves);
  GenCaptureKingMoves(pos, attackSide, attackOpp, pinned, moves);

  // could be regrouped for better perfs ?
  GenCaptureRookMoves(pos, attackSide, attackOpp, pinned, moves);
  GenCaptureBishopMoves(pos, attackSide, attackOpp, pinned, moves);
  GenCaptureQueenMoves(pos, attackSide, attackOpp, pinned, moves);

}

void GenCapturePawnMoves(Position& pos, attack_t& attackSide, attack_t& attackOpp, Bitboard& pinned, list_t& moves)
{
  Color side = pos.GetTurn();
  Color opp = (side ==WHITE) ? BLACK : WHITE;  
  int s = (pos.GetTurn()==WHITE) ? 0 : 8;
  Bitboard bbf, bbt;
  int fsq,tsq;

  bbf = pos.board.piecesBB[s+PAWN] & ~pinned;
  while (bbf)
  {
    fsq=pop_1st_bit(&bbf);
    bbt = StepAttacksBB[s+PAWN][fsq] & pos.board.colorBB[opp] & attackSide.pawnAttacks;
    while(bbt)
    {
      tsq=pop_1st_bit(&bbt);
      LIST_ADD(moves,make_move((Square)fsq,(Square)tsq));
    }
  }
}

void GenCaptureKnightMoves(Position& pos, attack_t& attackSide, attack_t& attackOpp, Bitboard& pinned, list_t& moves)
{
  Color side = pos.GetTurn();
  Color opp = (side ==WHITE) ? BLACK : WHITE;
  int s = (pos.GetTurn()==WHITE) ? 0 : 8;
  Bitboard bbf, bbt;
  int fsq,tsq;

  // knights
  bbf = pos.board.piecesBB[s+KNIGHT] & ~pinned;
  while (bbf)
  {
    fsq=pop_1st_bit(&bbf);
    bbt = StepAttacksBB[s+KNIGHT][fsq] & pos.board.colorBB[opp] & attackSide.knightAttacks;
    while(bbt)
    {
      tsq=pop_1st_bit(&bbt);
      LIST_ADD(moves,make_move((Square)fsq,(Square)tsq));
    }
  }
}

void GenCaptureKingMoves(Position& pos, attack_t& attackSide, attack_t& attackOpp, Bitboard& pinned, list_t& moves)
{
  Color side = pos.GetTurn();
  Color opp = (side ==WHITE) ? BLACK : WHITE;
  int s = (side==WHITE) ? 0 : 8;
  Bitboard bbf, bbt;
  int fsq,tsq;

  // King
  bbf = pos.board.piecesBB[s+KING];
  fsq=pop_1st_bit(&bbf);
  bbt = attackSide.kingAttacks & pos.board.colorBB[opp] & ~(attackOpp.allAttacks);
  while(bbt)
  {
    tsq=pop_1st_bit(&bbt);
    LIST_ADD(moves,make_move((Square)fsq,(Square)tsq));
  }
}

void GenCaptureRookMoves(Position& pos, attack_t& attackSide, attack_t& attackOpp, Bitboard& pinned, list_t& moves)
{
  //Color side = pos.GetTurn();
  Color opp = (pos.GetTurn()==WHITE) ? BLACK : WHITE;  
  int s = (pos.GetTurn()==WHITE) ? 0 : 8;
  Bitboard bbf, bbt;
  int fsq,tsq;

  // Rook
  bbf = pos.board.piecesBB[s+ROOK] & ~pinned;
  while (bbf)
  {
    fsq=pop_1st_bit(&bbf);
    bbt = PseudoAttacks[ROOK][fsq] & pos.board.colorBB[opp] & attackSide.rookAttacks;
    while(bbt)
    {
      tsq=pop_1st_bit(&bbt);
      LIST_ADD(moves,make_move((Square)fsq,(Square)tsq));
    }
  }

}

void GenCaptureBishopMoves(Position& pos, attack_t& attackSide, attack_t& attackOpp, Bitboard& pinned, list_t& moves)
{
  //Color side = pos.GetTurn();
  Color opp = (pos.GetTurn()==WHITE) ? BLACK : WHITE;  
  int s = (pos.GetTurn()==WHITE) ? 0 : 8;
  Bitboard bbf, bbt;
  int fsq,tsq;

  // Bishop
  bbf = pos.board.piecesBB[s+BISHOP] & ~pinned;
  while (bbf)
  {
    fsq=pop_1st_bit(&bbf);
    bbt = PseudoAttacks[BISHOP][fsq] & pos.board.colorBB[opp] & attackSide.bishopAttacks;
    //std::cout <<  "Bishop captures" << std::endl; print(bbt);
    while(bbt)
    {
      tsq=pop_1st_bit(&bbt);
      LIST_ADD(moves,make_move((Square)fsq,(Square)tsq));
    }
  }
}

void GenCaptureQueenMoves(Position& pos, attack_t& attackSide, attack_t& attackOpp, Bitboard& pinned, list_t& moves)
{
  //Color side = pos.GetTurn();
  Color opp = (pos.GetTurn()==WHITE) ? BLACK : WHITE;  
  int s = (pos.GetTurn()==WHITE) ? 0 : 8;
  Bitboard bbf, bbt;
  int fsq,tsq;

  // knights
  bbf = pos.board.piecesBB[s+QUEEN] & ~pinned;
  while (bbf)
  {
    fsq=pop_1st_bit(&bbf);
    bbt = PseudoAttacks[QUEEN][fsq] &  pos.board.colorBB[opp] & attackSide.queenAttacks;
    while(bbt)
    {
      tsq=pop_1st_bit(&bbt);
      LIST_ADD(moves,make_move((Square)fsq,(Square)tsq));
    }
  }
}



void GenQuietPawnMoves(Position& pos, attack_t& attackSide, attack_t& attackOpp, Bitboard& pinned, list_t& moves)
{
  Color side = pos.GetTurn();
  int s = (pos.GetTurn()==WHITE) ? 0 : 8;
  //int fw = (pos.GetTurn()==WHITE) ? 8 : -8;
  Bitboard bbf, bbt;
  int fsq,tsq;

  // pawn (generic code)
  bbf = pos.board.piecesBB[s+PAWN] & ~pinned;
  //print(bbf);
  while (bbf)
  {
    fsq=pop_1st_bit(&bbf);

    if(side == WHITE)
    {
      bbt = (SquareBB[fsq] << 8) & pos.board.emptyBB;
      bbt|= ((bbt & RankBB[2]) << 8) & pos.board.emptyBB;
    }
    else
    {
      bbt = (SquareBB[fsq] >> 8) & pos.board.emptyBB;
      bbt|= ((bbt & RankBB[5]) >> 8) & pos.board.emptyBB;
    }

    //std::cout <<  "square " << fsq << std::endl;
    //print(bbt);
    while(bbt)
    {
      tsq=pop_1st_bit(&bbt);
      if((rank_of((Square)tsq) == RANK_1) || (rank_of((Square)tsq) == RANK_8))
      {
        LIST_ADD(moves,make_promotion((Square)(fsq),(Square)tsq,QUEEN));
        LIST_ADD(moves,make_promotion((Square)(fsq),(Square)tsq,ROOK));
        LIST_ADD(moves,make_promotion((Square)(fsq),(Square)tsq,BISHOP));
        LIST_ADD(moves,make_promotion((Square)(fsq),(Square)tsq,KNIGHT));      
      }
      else
      {
        LIST_ADD(moves,make_move((Square)(fsq),(Square)tsq));
      }
    }
  }
}


void GenQuietKnightMoves(Position& pos, attack_t& attackSide, attack_t& attackOpp, Bitboard& pinned, list_t& moves)
{
  //Color side = pos.GetTurn();
  int s = (pos.GetTurn()==WHITE) ? 0 : 8;
  //int fw = (pos.GetTurn()==WHITE) ? 8 : -8;
  Bitboard bbf, bbt;
  int fsq,tsq;

  // knights
  bbf = pos.board.piecesBB[s+KNIGHT] & ~pinned;
  while (bbf)
  {
    fsq=pop_1st_bit(&bbf);
    bbt = StepAttacksBB[s+KNIGHT][fsq] & pos.board.emptyBB;
    while(bbt)
    {
      tsq=pop_1st_bit(&bbt);
      LIST_ADD(moves,make_move((Square)fsq,(Square)tsq));
    }
  }
}

void GenQuietKingMoves(Position& pos, attack_t& attackSide, attack_t& attackOpp, Bitboard& pinned, list_t& moves)
{
  //Color side = pos.GetTurn();
  int s = (pos.GetTurn()==WHITE) ? 0 : 8;
  //int fw = (pos.GetTurn()==WHITE) ? 8 : -8;
  Bitboard bbf, bbt;
  int fsq,tsq;

  // king
  bbf = pos.board.piecesBB[s+KING];
  fsq=pop_1st_bit(&bbf);
  bbt = StepAttacksBB[s+KING][fsq] & pos.board.emptyBB & ~attackOpp.allAttacks;
  while(bbt)
  {
    tsq=pop_1st_bit(&bbt);
    LIST_ADD(moves,make_move((Square)fsq,(Square)tsq));
  }
}

void GenQuietRookMoves(Position& pos, attack_t& attackSide, attack_t& attackOpp, Bitboard& pinned, list_t& moves)
{

  //Color side = pos.GetTurn();
  //Color opp = (pos.GetTurn()==WHITE) ? BLACK : WHITE;  
  int s = (pos.GetTurn()==WHITE) ? 0 : 8;
  Bitboard bbf, bbt;
  int fsq,tsq;

  // Rook
  bbf = pos.board.piecesBB[s+ROOK] & ~pinned;
  while (bbf)
  {
    fsq=pop_1st_bit(&bbf);
    bbt = PseudoAttacks[ROOK][fsq] & pos.board.emptyBB & attackSide.rookAttacks;
    while(bbt)
    {
      tsq=pop_1st_bit(&bbt);
      LIST_ADD(moves,make_move((Square)fsq,(Square)tsq));
    }
  }

}

void GenQuietBishopMoves(Position& pos, attack_t& attackSide, attack_t& attackOpp, Bitboard& pinned, list_t& moves)
{
  //Color side = pos.GetTurn();
  //Color opp = (pos.GetTurn()==WHITE) ? BLACK : WHITE;  
  int s = (pos.GetTurn()==WHITE) ? 0 : 8;
  Bitboard bbf, bbt;
  int fsq,tsq;

  // Bishop
  bbf = pos.board.piecesBB[s+BISHOP] & ~pinned;
  while (bbf)
  {
    fsq=pop_1st_bit(&bbf);
    bbt = PseudoAttacks[BISHOP][fsq] & pos.board.emptyBB & attackSide.bishopAttacks;
    //std::cout <<  "Bishop quiet moves" << std::endl; print(bbt);
    while(bbt)
    {
      tsq=pop_1st_bit(&bbt);
      LIST_ADD(moves,make_move((Square)fsq,(Square)tsq));
    }
  }

}

void GenQuietQueenMoves(Position& pos, attack_t& attackSide, attack_t& attackOpp, Bitboard& pinned, list_t& moves)
{
  //Color side = pos.GetTurn();
  //Color opp = (pos.GetTurn()==WHITE) ? BLACK : WHITE;  
  int s = (pos.GetTurn()==WHITE) ? 0 : 8;
  Bitboard bbf, bbt;
  int fsq,tsq;

  // knights
  bbf = pos.board.piecesBB[s+QUEEN] & ~pinned;
  while (bbf)
  {
    fsq=pop_1st_bit(&bbf);
    bbt = PseudoAttacks[QUEEN][fsq] &  pos.board.emptyBB & attackSide.queenAttacks;
    while(bbt)
    {
      tsq=pop_1st_bit(&bbt);
      LIST_ADD(moves,make_move((Square)fsq,(Square)tsq));
    }
  }  
}


void GenCastleMoves(Position& pos, attack_t& attackSide, attack_t& attackOpp, Bitboard& pinned, list_t& moves)
{
  //Color side = pos.GetTurn();
  //Color opp = (pos.GetTurn()==WHITE) ? BLACK : WHITE;  
  int s = (pos.GetTurn()==WHITE) ? 0 : 8;
  Bitboard bbf;//, bbt;
  int fsq;//,tsq;

  bbf = pos.board.piecesBB[s+KING] ;
  fsq=pop_1st_bit(&bbf);


  if (pos.GetTurn()==WHITE)
  {
    if (pos.gameState.whiteCanCastleQueenSide)
      if(pos.board.piecesBB[W_ROOK] & SquareBB[SQ_A1])
        if(pos.board.emptyBB & SquareBB[SQ_B1] )
          if(pos.board.emptyBB & SquareBB[SQ_C1] & ~attackOpp.allAttacks)
            if(pos.board.emptyBB & SquareBB[SQ_D1] & ~attackOpp.allAttacks)
              LIST_ADD(moves,make_move(SQ_E1, SQ_C1));
    if (pos.gameState.whiteCanCastleKingSide)
      if(pos.board.piecesBB[W_ROOK] & SquareBB[SQ_H1])
        if(pos.board.emptyBB & SquareBB[SQ_F1] & ~attackOpp.allAttacks)
          if(pos.board.emptyBB & SquareBB[SQ_G1] & ~attackOpp.allAttacks)
            LIST_ADD(moves,make_move(SQ_E1, SQ_G1));
  }
  else
  {
    if (pos.gameState.blackCanCastleQueenSide)
      if(pos.board.piecesBB[B_ROOK] & SquareBB[SQ_A8])
        if(pos.board.emptyBB & SquareBB[SQ_B8] )
          if(pos.board.emptyBB & SquareBB[SQ_C8] & ~attackOpp.allAttacks)
            if(pos.board.emptyBB & SquareBB[SQ_D8] & ~attackOpp.allAttacks)
              LIST_ADD(moves,make_move(SQ_E8, SQ_C8));
    if (pos.gameState.blackCanCastleKingSide)
      if(pos.board.piecesBB[B_ROOK] & SquareBB[SQ_H8])
        if(pos.board.emptyBB & SquareBB[SQ_F8] & ~attackOpp.allAttacks)
          if(pos.board.emptyBB & SquareBB[SQ_G8] & ~attackOpp.allAttacks)
            LIST_ADD(moves,make_move(SQ_E8, SQ_G8));
  }

}
*/

/*
// prune moves wich are not captures of checking moves
void futility_pruning(Position& pos, list_t& t)
{
  // copy elements and drop the rest of the list
  // no temporary vector 
  Color turn = pos.GetTurn();
  int lastLegalMoveIndex = 0;
  for(int i=0 ; i < t.size ; i++ )
  {
    Move m = t.move[i];
    if(!pos.IsEmpty(to_sq(m)))
    {
      if(i != lastLegalMoveIndex) // do the copy only if needed
        t.move[lastLegalMoveIndex] = m;
      lastLegalMoveIndex++;
    }
    else
    {
      pos.Play(m);
      if(pos.IsCheck())
      {
        if(i != lastLegalMoveIndex) // do the copy only if needed
          t.move[lastLegalMoveIndex] = m;
        lastLegalMoveIndex++;
      }
      pos.Unplay(m);
    }
  }
  // resize the list
  t.size = lastLegalMoveIndex;

}
*/

// search moves in pv and push them to the beginning of the list
void pv_reorder(list_t& moveList, map<Move,Move*> pvhash)
{
  int switchCursor = 0;
  for(int i=0;i<moveList.size;i++)
  {
    if(pvhash.count(LIST_MOVE(moveList,i)))
    {
      //LIST_SWITCH(moveList,i,switchCursor++);
      Move tm = LIST_MOVE(moveList,i);
      int ts = LIST_VALUE(moveList,i);
      LIST_MOVE(moveList,i) = LIST_MOVE(moveList,switchCursor);
      LIST_VALUE(moveList,i) = LIST_VALUE(moveList,switchCursor);
      LIST_MOVE(moveList,switchCursor) = tm;
      LIST_VALUE(moveList,switchCursor) = ts;
      switchCursor++;
    }
  }
}


