#ifndef POSITION_H
#define POSITION_H

#include <vector>
#include <string.h>
#include "Board.h"
#include "Piece.h"
#include "GameState.h"
using std::vector;

class Position
{

public:
	Board board; // should be private
	//Key zobrist;

public:
	Position();
	Position(Position & position);
	Position(string strPos);

	void SetFromFEN(string fenString);
	Color GetTurn();
  void SwitchTurn(); // illegal switch turn - for eval purpose only
	Piece GetPiece(Square square);
	char* GetSimplifiedHashCode();
	Key GetKey();     // get zobrist key

	bool IsCheck();
	bool IsCheck(Color side);
	bool IsCheck(Color side, Square square);
	bool IsPieceWhite(Piece piece);
	bool IsPieceBlack(Piece piece);
	bool IsEmptyOrOpponent(Square square);
	bool IsOpponent(Square square);
	bool IsEmpty(Square square);
	bool IsEmptyOrColor(Square square, bool color);

	void Play(Move m);
	void Unplay(Move m);
	void SetPiece(Piece piece, Square square);
	void SetStartPosition();


public: 
	GameState gameState;
	string GetFEN();
	bool IsEqual(Position& pos);
	
	vector<GameState> gameStateStack;
	void DebugDisplay();

	int Evaluate();
	int Evaluate(Color color);
	int GetPlayedMoveNumber();
	bool IsWhiteToMove();
	bool IsBlackToMove();

};


inline bool Position::IsWhiteToMove()
{
	return gameState.turn == WHITE;
}

inline bool Position::IsBlackToMove()
{
	return gameState.turn == BLACK;
}

inline bool Position::IsEmpty(Square square){
	return board.IsSquareEmpty(square);
}

inline Color Position::GetTurn()
{
	//return gameStateStack.back().intTurn;
	return gameState.turn;
}

inline void Position::SwitchTurn()
{
  gameState.turn = (gameState.turn==WHITE)?BLACK:WHITE;
}

// TODO: optimize this method
inline bool Position::IsPieceWhite(Piece piece)
{
	return (piece == W_PAWN || piece == W_ROOK || piece == W_KNIGHT
		|| piece == W_BISHOP || piece == W_QUEEN || piece == W_KING);
}

// TODO: optimize this method
inline bool Position::IsPieceBlack(Piece piece)
{
	return (piece == B_PAWN || piece == B_ROOK || piece == B_KNIGHT
		|| piece == B_BISHOP || piece == B_QUEEN || piece == B_KING);
}

inline bool Position::IsEmptyOrOpponent(Square square)
{
  Color opp = (GetTurn() == WHITE) ? BLACK: WHITE;
  return (board.colorBB[opp] | board.emptyBB) & SquareBB[square] ;
}

inline bool Position::IsOpponent(Square square)
{
  Color opp = (GetTurn() == WHITE) ? BLACK: WHITE;
  return board.colorBB[opp] & SquareBB[square] ;
}

inline bool Position::IsEmptyOrColor(Square square, bool color)
{
  return (board.colorBB[color] | board.emptyBB) & SquareBB[square] ;
}

inline Piece Position::GetPiece(Square square)
{
	return board.GetPiece(square);
}

#endif
