#include <assert.h>
#include "Engine.h"
#include "Search.h"
#include "Zobrist.h"
#include "Transposition.h"



void pv_copy(const Move* src, Move* dst, int depth, int maxdepth)
{
  for(int i=depth;i<=(maxdepth+1);i++) dst[i] = src[i];
}


void Search(Position& pos, int depth)
{
  int value,best_value;
  selectiveDepth = depth ;
  //selectiveDepth = (int)(1.5*depth) ;//arbitrary selective depth;
  nbProcessedNodes++;
  nbCurrSearchTotalGenNodes=0;
  nbCurrSearchProcessedNodes=0;
 
  //std::cout <<  "SEARCH LAUNCHED at depth " << depth << std::endl;
  // TODO : global variables should be declared as local for multithreading

  // TODO: adjust alpha and beta values at level zero
  // alpha & beta are updated outside this function before the first call
  int alpha = -VALUE_INFINITE;
  int beta  = +VALUE_INFINITE;

  //int bestMoveIndex;
  //int ipv=0;
  //nbNodes+=moves.size;
  int old_alpha = alpha;
  best_value = VALUE_NONE;
  nbCurrSearchTotalGenNodes += LIST_SIZE(moves);
  for(int i=0;i<moves.size;i++)
  {
    Move currmove = LIST_MOVE(moves,i);
    Move* pv      = pvhash[currmove] = pvbuffer[i];
    pv[0]         = currmove;   
    int scDelta;

    // Set aspiration window default width
    // TODO: debug aspiration window (in conflict with ttable)
    if (depth >= 5 && abs(LIST_VALUE(moves,i)) < VALUE_KNOWN_WIN)
    {
      scDelta = 16;//Score(16); // from stockfish
        alpha = LIST_VALUE(moves,i) - scDelta;
        beta  = LIST_VALUE(moves,i) + scDelta;
    }
    else
    {
        alpha = -VALUE_INFINITE;
        beta  = +VALUE_INFINITE;
    }

    pos.Play(currmove);
    nbCurrSearchProcessedNodes ++;
    if(best_value == VALUE_NONE)
    {
      // first move, full-window search
      value = -full_search(pos, -beta, -alpha, 1, depth, pv);
    }
    else
    {
      // not the first move, zero-window search
      value = -full_search(pos, -alpha-1, -alpha, 1, depth, pv);
      if(value > alpha)
      {
        //.. && score < beta
        // need to perform a full-window re-search
        value = -full_search(pos, -beta, -alpha, 1, depth, pv);
      }
    }
    
    
    
    /*
    while(true)
    {
      score = -Negmax(pos, alpha, beta, 1, depth, pv);
      //score = -pvSearch(pos, alpha, beta, 1, depth, pv);
      //if(brainStatus != THINKING) break;
      if(brainStatus != THINKING) return;
      
      // In case of failing high/low increase aspiration window and
      // research, otherwise exit the fail high/low loop.
      if (score >= beta)
      {
          beta += scDelta;
          scDelta += scDelta / 2;
      }
      else if (score <= alpha)
      {
          alpha -= scDelta;
          scDelta += scDelta / 2;
      }
      else
          break;
      // Search with full window in case we have a win/mate score
      if (abs(score) >= VALUE_KNOWN_WIN)
      {
        alpha = -VALUE_INFINITE;
        beta  = +VALUE_INFINITE;
      }
    }
    */
    pos.Unplay(currmove);
    
    if(value <= alpha) // upper bound
    {
      LIST_VALUE(moves,i) = old_alpha;
    }
    else if(value >= beta) // lower bound
    {
      LIST_VALUE(moves,i) = beta;
    }
    else // alpha < value < beta (exact value)
    {
      LIST_VALUE(moves,i) = value;
    }
    
    
    if(value>best_value && (best_value == VALUE_NONE || value > alpha))
    {
      bestMove = currmove;
      bestScore = value;
      
      /*
      if(value <= alpha) // upper bound
      {
        // set flag searchUpper
      }
      else if(value >= beta) // lower bound
      {
        // set flag searchLower
      }
      else // alpha < value < beta (exact value)
      {
        // set flag searchExact
      }
      */
    }
    
    if(value > best_value)
    {
      if(value>alpha)
      {
          //if(search_type == SearchNormal) 
            alpha = value;
          if(value >= beta) break;
      }
    
    }


    // after the 'while' loop the score is retrieved with the correct search window
    // update only if the search was not aborted
    //LIST_SCORE(moves,i) = score;
  }
  
  // TODO: handle result properly if (brainStatus != THINKING)
  // score might not be correct when returning like this
  // TODO: store previous score
  if(brainStatus != THINKING) return;

  // At this point all the moves are properly evaluated
  // sort moves by score
  Sort(moves);

  // TODO: implement UCI_LimitStrength directly in the eval function to induce some erroneous estimation
  // no limitation: best score at index 0
  int bestMoveIndex = 0;
  bestMove  = LIST_MOVE(moves,bestMoveIndex);
  bestScore = LIST_VALUE(moves,bestMoveIndex);
  bestPv    = pvhash[bestMove]; 
  DisplayBestPv(depth, nbProcessedNodes, watch.GetElapsedTime(), bestPv);
  if(debugMode) DisplayScores(moves, depth);
  //std::cout <<  "### bestmove=" << move_to_string(bestMove) << "  bestscore=" << bestScore <<  std::endl;
}


void display_move_buffer(Move* moveBuff, int depth)
{
  std::cout <<  "MOVE BUFFER: ";
  for(int i=0;i<depth;i++)
  {
    std::cout <<  move_to_string(moveBuff[i]) << " ";
  }
  std::cout <<  std::endl;
}

void display_board_buffer(Move* moveBuff, Board* boardBuff, int depth)
{
  for(int i=0;i<depth;i++)
  {
    std::cout <<  "AFTER PLAYING MOVE " << move_to_string(moveBuff[i]) << std::endl;
    boardBuff[i].DebugDisplay();
  }
}

// Self-called: iso-AlphaBeta (2 fct implementation in one)
// uses Hash table
int full_search(Position& pos, int alpha, int beta, int depth, int maxDepth, Move* pv)
{
  int value     = VALUE_NONE;
  int bestValue = VALUE_NONE;
  nbProcessedNodes++;
  if (brainStatus != THINKING) return VALUE_NONE; // quick exit
 
//  const int TransDepth = 2; 
//  Key key = book_key(pos);
  Move localPv[MAX_MOVES];
  Move localBestPv[MAX_MOVES];
  int hashvalue;
  
  
  if(useTrans)
    if(TTable.lookup(&pos, depth, (maxDepth - depth), &alpha, &beta, &hashvalue))
		  return hashvalue;

  if(depth >= selectiveDepth)
  //if (depth == maxDepth)
  {
    // last move
    pv[depth] = MOVE_NULL;
    return EvaluateBoard(pos.board,GetTurn());
  }

  list_t localmoves;
  GetLegalMoves(pos, localmoves);
  if (LIST_IS_EMPTY(localmoves))
  {
    // last move
    pv[depth] = MOVE_NULL;
    if (pos.IsCheck())
    {
      return (pos.GetTurn() != GetTurn()) ? mate_in(depth) : mated_in(depth);
    }
    else
    {
      return VALUE_DRAW;
    }
  }

  /*
  // Selectivity pruning
  if(depth > maxDepth)
  {
    SelectivityPruning(pos, localmoves); 
  }
  // TODO: in case of empty list return the correct result
  */

  //Move bestm = moves.move[0];// there is at least 1 valid move (see LIST_IS_EMPTY upper statment )
  localPv[depth] = localmoves.move[0];
  bestValue = VALUE_NONE;
  //nbNodes+=moves.size;
  nbCurrSearchTotalGenNodes += localmoves.size;
  // optimistic value for futility pruning
  int opt_value = +VALUE_INFINITE;
  for(int i=0;i<localmoves.size;i++)
  {
    // extensions (augment depth if needed)
    
    // history pruning
    
    // futility pruning
    bool UseFutility = true;
    int FutilityMargin = 100;
    // TODO: check if 'futile' move is not a pawn pushed on the 7th rank
    if(UseFutility && depth == (maxDepth-1))
    {
      // TODO: apply alpha-beta on pre-leave nodes using futilityMargin
      if(opt_value == +VALUE_INFINITE)
      {
        opt_value = EvaluateBoard(pos.board,GetTurn())+FutilityMargin;
      }
      value = opt_value;
      // pruning
      if(value<=alpha)
      {
        if(value>bestValue)
        {
          bestValue = value;
        }
        // jump to the next move
        continue;
      }
    }
    
    Move m = localmoves.move[i];
    pos.Play(m);
    nbCurrSearchProcessedNodes++;
    if(bestValue == VALUE_NONE) // first move, always search with full window
    {
      value = -full_search(pos, -beta, -alpha, depth + 1, maxDepth, localPv);
    }
    else // not the first move, zero window search first
    {
      value = -full_search(pos, -alpha-1, -alpha, depth + 1, maxDepth, localPv);
      if(value > alpha) // search failed, re-search with full window
        value = -full_search(pos, -beta, -alpha, depth + 1, maxDepth, localPv);
    }

    pos.Unplay(m);
    if(value >= beta)
    {
      //if(isNonCapture(m))
      if(pos.IsEmpty(to_sq(m)))
      {
        history[pos.GetPiece(from_sq(m))][to_sq(m)] += depth*depth;
        //if(debugMode) cout << "beta-cut-off(" << i << ")" << move_to_string(m) << " ";
        //if(debugMode) cout << "(" << i << ")";
      }
      // fail hard beta-cutoff - i.e. no pv move detected
      if( useTrans ) TTable.store(&pos, depth, (maxDepth - depth), value, alpha, beta);
      return beta;
    }
    if(value > alpha)
    {
      // alpha acts like max in MiniMax
      alpha = value;
      localPv[depth]=m; // keeps potential pv move
      pv_copy(localPv, localBestPv, depth, maxDepth); // Keeps potential best pv     
    }
  }
  
  if(useTrans && brainStatus == THINKING) TTable.store(&pos, depth, (maxDepth - depth), value, alpha, beta);
  
  pv_copy(localBestPv, pv, depth, maxDepth);
  return alpha;
}

/*
void SelectivityPruning(Position&pos, list_t& localmoves)
{
  // prune non-capture and non-checking moves
  Color oppTurn = BLACK-pos.GetTurn();
  int lastGoodMoveIndex = 0;
  for(int i=0 ; i < localmoves.size ; i++ )
  {
    Move m = localmoves.move[i];
    pos.Play(m);
    if(!pos.IsEmpty(to_sq(m)) || pos.IsCheck(oppTurn))
    {
      if(i != lastGoodMoveIndex) // do the copy only if needed
        localmoves.move[lastGoodMoveIndex] = m;
      lastGoodMoveIndex++;		// the move is good, always increment
    }
    pos.Unplay(m);
  }
  // resize the list
  localmoves.size = lastGoodMoveIndex;
} 
*/
