#include <assert.h>
#include "Zobrist.h"
#include "Transposition.h"


TranspositionTable::TranspositionTable() 
{
  ResizeMB(1); // 1 Mo by default
}

TranspositionTable::~TranspositionTable() 
{
}

void TranspositionTable::ResizeMB(int sizeInMB)
{
  assert(sizeInMB>0);
  //std::cout <<  "Resizing Transposition Table to " << sizeInMB "Mb" << std::endl; 

  // flush table content
  this->clear();
  // convert MB in bytes
  this->maxSize = (1048576L * sizeInMB) / (sizeof(TTElem)+sizeof(Key));
}


void TranspositionTable::clear()
{
  if(!table.empty())
    table.erase ( table.begin(), table.end() );
  hits = 0;
}


bool TranspositionTable::lookup(Position *p, int depth, int height, int* alpha, int* beta, int* hashvalue)
{
  map<Key, TTElem>::iterator result;
  result = table.find(p->GetKey());
  if(result==table.end()) 
    return false; // no entry found
  //if(result->second.depth < depth) 
  if(result->second.depth < height) 
    return false; 
  
  // entry found, update values
 if((depth)%2)
  {
    //std::cout <<  "WARNING !!! - odd hashtable hit" << std::endl; 
    (*alpha)      = -result->second.beta;
    (*beta)       = -result->second.alpha;
    (*hashvalue)  = -result->second.bestvalue;
  }
  else{
 
    //std::cout <<  "regular hashtable hit" << std::endl; 
    (*alpha)      = result->second.alpha;
    (*beta)       = result->second.beta;
    (*hashvalue)  = result->second.bestvalue;
  }
  
  hits++;
  return true; 
}


void TranspositionTable::store(Position *p, int depth, int height, int bestvalue, int alpha, int beta)
{
  Key key = p->GetKey();
  map<Key, TTElem>::iterator result;
  result = table.find(key);
  if(result==table.end()) 
  {
    // no entry, always store
    //TTElem elem = {depth,bestvalue,alpha,beta};
    TTElem elem = {height,bestvalue,alpha,beta};
    table[key] = elem;
    if(table.size() > maxSize) table.erase(table.begin());
  }
  else
  {
    // entry found, update only deeper results
    if(height > result->second.depth) 
    {
      //result->second.depth = depth;
      result->second.depth = height;
      if((depth)%2)
      {
        result->second.alpha = -beta;
        result->second.beta  = -alpha;
        result->second.bestvalue = -bestvalue;
      }
      else
      {
        result->second.alpha = alpha;
        result->second.beta  = beta;
        result->second.bestvalue = bestvalue;
      } 
    }
  }
}


unsigned int TranspositionTable::GetUsagePerMil()
{
  return  (1000L * (unsigned long)table.size()) / maxSize;
}


unsigned long TranspositionTable::GetHits()
{
  return hits;
}


    
    

















